"""Functions for caper init subcommand

Author:
    Jin Lee (leepc12@gmail.com) at ENCODE-DCC
"""

import os
import sys
from autouri import AutoURI, AbsPath
from .caper_backend import BACKENDS, BACKENDS_WITH_ALIASES
from .caper_backend import BACKEND_GCP, BACKEND_AWS, BACKEND_LOCAL
from .caper_backend import BACKEND_SLURM, BACKEND_SGE, BACKEND_PBS
from .caper_backend import BACKEND_ALIAS_LOCAL
from .caper_backend import BACKEND_ALIAS_GOOGLE, BACKEND_ALIAS_AMAZON
from .caper_backend import BACKEND_ALIAS_SHERLOCK, BACKEND_ALIAS_SCG
from .caper_args import DEFAULT_CROMWELL_JAR, DEFAULT_WOMTOOL_JAR


DEFAULT_CROMWELL_JAR_INSTALL_DIR = '~/.caper/cromwell_jar'
DEFAULT_WOMTOOL_JAR_INSTALL_DIR = '~/.caper/womtool_jar'
DEFAULT_CONF_CONTENTS_LOCAL = """backend=local

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""
DEFAULT_CONF_CONTENTS_SHERLOCK = """backend=slurm
slurm-partition=

# DO NOT use /tmp here
# You can use $OAK or $SCRATCH storages here.
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=

# IMPORTANT warning for Stanford Sherlock cluster
# ====================================================================
# DO NOT install any codes/executables
# (java, conda, python, caper, pipeline's WDL, pipeline's Conda env, ...) on $SCRATCH or $OAK.
# You will see Segmentation Fault errors.
# Install all executables on $HOME or $PI_HOME instead.
# It's STILL OKAY to read input data from and write outputs to $SCRATCH or $OAK.
# ====================================================================
"""
DEFAULT_CONF_CONTENTS_SCG = """backend=slurm
slurm-account=

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""
DEFAULT_CONF_CONTENTS_SLURM = """backend=slurm

# define one of the followings (or both) according to your
# cluster's SLURM configuration.
slurm-partition=
slurm-account=

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""
DEFAULT_CONF_CONTENTS_SGE = """backend=sge
sge-pe=

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""
DEFAULT_CONF_CONTENTS_PBS = """backend=pbs

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""
DEFAULT_CONF_CONTENTS_AWS = """backend=aws
aws-batch-arn=
aws-region=
out-s3-bucket=

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""
DEFAULT_CONF_CONTENTS_GCP = """backend=gcp
gcp-prj=
out-gcs-bucket=

# call-cached outputs will be duplicated by making a copy or reference
#  reference: refer to old output file in metadata.json file.
#  copy: make a copy
gcp-call-caching-dup-strat=

# DO NOT use /tmp here
# Caper stores all important temp files and cached big data files here
# If not defined, Caper will make .caper_tmp/ on your local output directory
# which is defined by out-dir, --out-dir or $CWD
# Use a local absolute path here
tmp-dir=
"""


def install_cromwell_jar(uri):
    """Download cromwell-X.jar
    """
    u = AutoURI(uri)
    if isinstance(u, AbsPath):
        return u.uri
    print('Downloading Cromwell JAR... {f}'.format(f=uri), file=sys.stderr)
    path = os.path.join(
        os.path.expanduser(DEFAULT_CROMWELL_JAR_INSTALL_DIR),
        os.path.basename(uri))
    return u.cp(path)


def install_womtool_jar(uri):
    """Download womtool-X.jar
    """
    u = AutoURI(uri)
    if isinstance(u, AbsPath):
        return u.uri
    print('Downloading Womtool JAR... {f}'.format(f=uri), file=sys.stderr)
    path = os.path.join(
        os.path.expanduser(DEFAULT_WOMTOOL_JAR_INSTALL_DIR),
        os.path.basename(uri))
    return u.cp(path)


def init_caper_conf(args):
    """Initialize conf file for a given platform.
    Also, download/install Cromwell/Womtool JARs.
    """
    backend = args.get('platform')
    assert(backend in BACKENDS_WITH_ALIASES)
    if backend in (BACKEND_LOCAL, BACKEND_ALIAS_LOCAL):
        contents = DEFAULT_CONF_CONTENTS_LOCAL
    elif backend == BACKEND_ALIAS_SHERLOCK:
        contents = DEFAULT_CONF_CONTENTS_SHERLOCK
    elif backend == BACKEND_ALIAS_SCG:
        contents = DEFAULT_CONF_CONTENTS_SCG
    elif backend == BACKEND_SLURM:
        contents = DEFAULT_CONF_CONTENTS_SLURM
    elif backend == BACKEND_SGE:
        contents = DEFAULT_CONF_CONTENTS_SGE
    elif backend == BACKEND_PBS:
        contents = DEFAULT_CONF_CONTENTS_PBS
    elif backend in (BACKEND_GCP, BACKEND_ALIAS_GOOGLE):
        contents = DEFAULT_CONF_CONTENTS_GCP
    elif backend in (BACKEND_AWS, BACKEND_ALIAS_AMAZON):
        contents = DEFAULT_CONF_CONTENTS_AWS
    else:
        raise Exception('Unsupported platform/backend/alias.')

    conf_file = os.path.expanduser(args.get('conf'))
    with open(conf_file, 'w') as fp:
        fp.write(contents + '\n')
        fp.write('{key}={val}\n'.format(
            key='cromwell',
            val=install_cromwell_jar(DEFAULT_CROMWELL_JAR)))
        fp.write('{key}={val}\n'.format(
            key='womtool',
            val=install_womtool_jar(DEFAULT_WOMTOOL_JAR)))



"""CaperCheck: Caper arguments/configuration checker

Author:
    Jin Lee (leepc12@gmail.com) at ENCODE-DCC
"""

import os
from .caper_backend import BACKENDS, BACKEND_SLURM, get_backend

DEFAULT_FILE_DB_PREFIX = 'caper_file_db'
DEFAULT_CAPER_TMP_DIR_SUFFIX = '.caper_tmp'


def check_caper_conf(args_d):
    """Check arguments/configuration for Caper
    """
    backend = args_d.get('backend')
    if backend is not None:
        backend = get_backend(backend)
    args_d['backend'] = backend

    docker = args_d.get('docker')
    if docker is not None:
        use_docker = True
    else:
        use_docker = False
    if isinstance(docker, list):
        if len(docker) > 0:
            docker = docker[0]
        else:
            docker = None
    elif isinstance(docker, str) and docker == '':
        docker = None
    args_d['docker'] = docker
    args_d['use_docker'] = use_docker

    singularity = args_d.get('singularity')
    if singularity is not None:
        use_singularity = True
    else:
        use_singularity = False
    if isinstance(singularity, list):
        if len(singularity) > 0:
            singularity = singularity[0]
        else:
            singularity = None
    elif isinstance(singularity, str) and singularity == '':
        singularity = None
    args_d['singularity'] = singularity
    args_d['use_singularity'] = use_singularity

    if use_docker and use_singularity:
        raise Exception('--docker and --singularity are mutually exclusive')

    # init some important path variables
    if args_d.get('out_dir') is None:
        args_d['out_dir'] = os.getcwd()
    else:
        args_d['out_dir'] = os.path.abspath(
            os.path.expanduser(args_d['out_dir']))

    if args_d.get('tmp_dir') is None:
        args_d['tmp_dir'] = os.path.abspath(
            os.path.join(args_d['out_dir'], DEFAULT_CAPER_TMP_DIR_SUFFIX))
    else:
        args_d['tmp_dir'] = os.path.abspath(
            os.path.expanduser(args_d['tmp_dir']))

    if args_d.get('tmp_s3_bucket') is None:
        if args_d.get('out_s3_bucket'):
            args_d['tmp_s3_bucket'] = os.path.join(args_d['out_s3_bucket'],
                                                   DEFAULT_CAPER_TMP_DIR_SUFFIX)
    if args_d.get('tmp_gcs_bucket') is None:
        if args_d.get('out_gcs_bucket'):
            args_d['tmp_gcs_bucket'] = os.path.join(args_d['out_gcs_bucket'],
                                                    DEFAULT_CAPER_TMP_DIR_SUFFIX)
    file_db = args_d.get('file_db')
    if file_db is not None:
        file_db = os.path.abspath(os.path.expanduser(file_db))
    else:
        basename = DEFAULT_FILE_DB_PREFIX
        out_dir = args_d['out_dir']
        inputs = args_d.get('inputs')
        if inputs is not None:
            inputs_wo_ext, ext = os.path.splitext(
                os.path.basename(inputs))
            basename += '.' + inputs_wo_ext
        file_db = os.path.join(out_dir, basename)
    args_d['file_db'] = file_db

    singularity_cachedir = args_d.get('singularity_cachedir')
    if singularity_cachedir is not None:
        singularity_cachedir = os.path.abspath(
            os.path.expanduser(singularity_cachedir))
        args_d['singularity_cachedir'] = singularity_cachedir
        os.makedirs(singularity_cachedir, exist_ok=True)

    if args_d.get('str_label') is None:
        if args_d.get('inputs') is not None:
            basename = os.path.basename(args_d['inputs'])
            args_d['str_label'] = os.path.splitext(basename)[0]

    if backend == 'sge':
        if args_d.get('sge_pe') is None:
            raise Exception(
                '--sge-pe (Sun GridEngine parallel environment) '
                'is required for backend sge.')
    elif backend =='gcp':
        if args_d.get('gcp_prj') is None:
            raise Exception(
                '--gcp-prj (Google Cloud Platform project) '
                'is required for backend gcp.')
        if args_d.get('out_gcs_bucket') is None:
            raise Exception(
                '--out-gcs-bucket (gs:// output bucket path) '
                'is required for backend gcp.')
    elif backend =='aws':
        if args_d.get('aws_batch_arn') is None:
            raise Exception(
                '--aws-batch-arn (ARN for AWS Batch) '
                'is required for backend aws.')
        if args_d.get('aws_region') is None:
            raise Exception(
                '--aws-region (AWS region) '
                'is required for backend aws.')
        if args_d.get('out_s3_bucket') is None:
            raise Exception(
                '--out-s3-bucket (s3:// output bucket path) '
                'is required for backend aws.')

    if args_d.get('soft_glob_output') and args_d.get('use_docker'):
        raise ValueError(
                '--soft-glob-output and --docker are mutually exclusive. '
                'Delocalization from docker container will fail '
                'for soft-linked globbed outputs.')

    return args_d



#!/usr/bin/env python3
"""Caper: Cromwell/WDL wrapper python script
for multiple backends (local, gc, aws)

(Optional)
Add the following comments to your WDL script to specify container images
that Caper will use for your WDL. You still need to add
"--use-docker" or "--use-singularity" to command line arguments.

Example:
#CAPER docker quay.io/encode-dcc/atac-seq-pipeline:v1.1.7.2
#CAPER singularity docker://quay.io/encode-dcc/atac-seq-pipeline:v1.1.7.2

Author:
    Jin Lee (leepc12@gmail.com) at ENCODE-DCC
"""

from pyhocon import ConfigFactory, HOCONConverter
import os
import sys
import logging
import pwd
import json
import re
import time
import socket
from threading import Thread
import shutil
from subprocess import Popen, check_call, PIPE, CalledProcessError
from datetime import datetime
from autouri import AutoURI, AbsPath, GCSURI, S3URI, URIBase
from autouri import logger as autouri_logger
from autouri.loc_aux import recurse_json
from tempfile import TemporaryDirectory
from .dict_tool import merge_dict
from .caper_args import parse_caper_arguments
from .caper_init import init_caper_conf, install_cromwell_jar, install_womtool_jar

from .caper_wdl_parser import CaperWDLParser
from .caper_check import check_caper_conf
from .cromwell_rest_api import CromwellRestAPI
from .caper_backend import BACKEND_GCP, BACKEND_AWS, BACKEND_LOCAL, \
    CaperBackendCommon, CaperBackendDatabase, CaperBackendGCP, \
    CaperBackendAWS, CaperBackendLocal, CaperBackendSLURM, \
    CaperBackendSGE, CaperBackendPBS
from .caper_backend import CaperBackendBase, CaperBackendBaseLocal


logging.basicConfig(level=logging.INFO, format='%(asctime)s|%(name)s|%(levelname)s| %(message)s')
logger = logging.getLogger('caper')


class Caper(object):
    """Cromwell/WDL wrapper
    """

    BACKEND_CONF_HEADER = 'include required(classpath("application"))\n'
    DEFAULT_BACKEND = BACKEND_LOCAL
    RE_PATTERN_BACKEND_CONF_HEADER = r'^\s*include\s'
    RE_PATTERN_STARTED_WORKFLOW_ID = \
        r'started WorkflowActor-(\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b)'
    RE_PATTERN_FINISHED_WORKFLOW_ID = \
        r'WorkflowActor-(\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b) is in a terminal state'
    RE_PATTERN_STARTED_CROMWELL_SERVER = \
        r'Cromwell \d+ service started on'
    RE_PATTERN_WORKFLOW_ID = \
        r'\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b'
    RE_PATTERN_DELIMITER_GCP_ZONES = r',| '
    USER_INTERRUPT_WARNING = '\n********** DO NOT CTRL+C MULTIPLE TIMES **********\n'

    MAX_RETRY_UPDATING_METADATA = 3
    SEC_INTERVAL_UPDATE_METADATA = 1200.0
    SEC_INTERVAL_UPDATE_SERVER_HEARTBEAT = 60.0
    SEC_INTERVAL_RETRY_UPDATING_METADATA = 10.0

    # added to cromwell labels file
    KEY_CAPER_STR_LABEL = 'caper-str-label'
    KEY_CAPER_USER = 'caper-user'
    KEY_CAPER_BACKEND = 'caper-backend'
    TMP_FILE_BASENAME_METADATA_JSON = 'metadata.json'
    TMP_FILE_BASENAME_WORKFLOW_OPTS_JSON = 'workflow_opts.json'
    TMP_FILE_BASENAME_BACKEND_CONF = 'backend.conf'
    TMP_FILE_BASENAME_LABELS_JSON = 'labels.json'
    TMP_FILE_BASENAME_IMPORTS_ZIP = 'imports.zip'
    COMMON_ROOT_SEARCH_LEVEL = 5  # to find common roots of files for singularity_bindpath

    def __init__(self, args):
        """Initializes from args dict
        """
        self._dry_run = args.get('dry_run')

        # init REST API
        self.__init_cromwell_rest_api(
            action=args.get('action'),
            ip=args.get('ip'),
            port=args.get('port'),
            no_server_hearbeat=args.get('no_server_heartbeat'),
            server_hearbeat_file=args.get('server_heartbeat_file'),
            server_hearbeat_timeout=args.get('server_heartbeat_timeout'))

        # java heap size
        self._java_heap_server = args.get('java_heap_server')
        self._java_heap_run = args.get('java_heap_run')

        # init others
        # self._keep_temp_backend_file = args.get('keep_temp_backend_file')
        self._hold = args.get('hold')
        self._format = args.get('format')
        self._hide_result_before = args.get('hide_result_before')
        self._disable_call_caching = args.get('disable_call_caching')
        self._max_concurrent_workflows = args.get('max_concurrent_workflows')
        self._max_concurrent_tasks = args.get('max_concurrent_tasks')
        self._max_retries = args.get('max_retries')
        self._tmp_dir = args.get('tmp_dir')
        self._out_dir = args.get('out_dir')
        if self._out_dir is not None:
            self._out_dir = os.path.abspath(self._out_dir)
        if self._tmp_dir is not None:
            self._tmp_dir = os.path.abspath(self._tmp_dir)
        self._gcp_prj = args.get('gcp_prj')
        self._gcp_zones = args.get('gcp_zones')
        self._gcp_call_caching_dup_strat = args.get('gcp_call_caching_dup_strat')
        self._out_gcs_bucket = args.get('out_gcs_bucket')
        self._out_s3_bucket = args.get('out_s3_bucket')
        self._aws_batch_arn = args.get('aws_batch_arn')
        self._aws_region = args.get('aws_region')
        self._slurm_partition = args.get('slurm_partition')
        self._slurm_account = args.get('slurm_account')
        self._slurm_extra_param = args.get('slurm_extra_param')
        self._sge_pe = args.get('sge_pe')
        self._sge_queue = args.get('sge_queue')
        self._sge_extra_param = args.get('sge_extra_param')
        self._pbs_queue = args.get('pbs_queue')
        self._pbs_extra_param = args.get('pbs_extra_param')

        self._backend_file = AbsPath.get_abspath_if_exists(
            args.get('backend_file'))
        self._soft_glob_output = args.get('soft_glob_output')
        self._wdl = AbsPath.get_abspath_if_exists(
            args.get('wdl'))
        self._inputs = AbsPath.get_abspath_if_exists(
            args.get('inputs'))
        self._cromwell = AbsPath.get_abspath_if_exists(
            args.get('cromwell'))
        self._workflow_opts = AbsPath.get_abspath_if_exists(
            args.get('options'))
        self._str_label = args.get('str_label')
        self._labels = AbsPath.get_abspath_if_exists(
            args.get('labels'))
        self._imports = AbsPath.get_abspath_if_exists(
            args.get('imports'))
        self._singularity_cachedir = args.get('singularity_cachedir')
        self._ignore_womtool = args.get('ignore_womtool')
        self._womtool = AbsPath.get_abspath_if_exists(
            args.get('womtool'))

        self._metadata_output = args.get('metadata_output')
        if self._metadata_output is not None and not AutoURI(self._metadata_output).is_valid:
            # metadata output doesn't exist at this moment since it's an output
            # so cannot use AbsPath.get_abspath_if_exists() here
            # make it abspath if it's given as local relative path ($CWD/relpath)
            self._metadata_output = os.path.abspath(self._metadata_output)

        # DB
        self._db = args.get('db')
        self._db_timeout = args.get('db_timeout')
        self._file_db = args.get('file_db')
        self._mysql_db_ip = args.get('mysql_db_ip')
        self._mysql_db_port = args.get('mysql_db_port')
        self._mysql_db_user = args.get('mysql_db_user')
        self._mysql_db_password = args.get('mysql_db_password')
        self._mysql_db_name = args.get('mysql_db_name')
        self._postgresql_db_ip = args.get('postgresql_db_ip')
        self._postgresql_db_port = args.get('postgresql_db_port')
        self._postgresql_db_user = args.get('postgresql_db_user')
        self._postgresql_db_password = args.get('postgresql_db_password')
        self._postgresql_db_name = args.get('postgresql_db_name')

        # troubleshoot
        self._show_completed_task = args.get('show_completed_task')

        # backend and default backend
        self._backend = args.get('backend')
        if self._backend is None:
            if args.get('action') == 'submit':
                self._backend = self._cromwell_rest_api.get_default_backend()
            else:
                self._backend = Caper.DEFAULT_BACKEND
        if self._backend == 'local':
            self._backend = BACKEND_LOCAL  # Local (capital L)

        # deepcopy
        self._no_deepcopy = args.get('no_deepcopy')

        # containers
        self._no_build_singularity = args.get('no_build_singularity')
        self._use_docker = args.get('use_docker')
        self._use_singularity = args.get('use_singularity')
        self._singularity = args.get('singularity')
        self._docker = args.get('docker')

        # list of values
        self._wf_id_or_label = args.get('wf_id_or_label')

    def run(self):
        """Run a workflow using Cromwell run mode
        """

        timestamp = Caper.__get_time_str()
        # otherwise, use WDL basename
        suffix = os.path.join(
            self.__get_wdl_basename_wo_ext(), timestamp)
        tmp_dir = self.__mkdir_tmp_dir(suffix)

        # all input files
        backend_file = self.__create_backend_conf_file(tmp_dir)
        input_file = self.__create_input_json_file(tmp_dir)
        workflow_opts_file = self.__create_workflow_opts_json_file(
            input_file, tmp_dir)
        labels_file = self.__create_labels_json_file(tmp_dir)

        if self._imports is not None:
            imports_file = AbsPath.localize(self._imports)
        else:
            imports_file = None
        wdl = AbsPath.localize(self._wdl)

        # metadata JSON file is an output from Cromwell
        #   place it on the tmp dir
        metadata_file = os.path.join(
            tmp_dir, Caper.TMP_FILE_BASENAME_METADATA_JSON)

        java_heap = '-Xmx{}'.format(self._java_heap_run)
        # LOG_LEVEL must be >=INFO to catch workflow ID from STDOUT
        cmd = ['java', java_heap, '-XX:ParallelGCThreads=1', '-DLOG_LEVEL=INFO',
               '-DLOG_MODE=standard',
               '-jar', '-Dconfig.file={}'.format(backend_file),
               install_cromwell_jar(self._cromwell), 'run',
               wdl,
               '-i', input_file,
               '-o', workflow_opts_file,
               '-l', labels_file,
               '-m', metadata_file]
        if imports_file is not None:
            cmd += ['-p', imports_file]

        self.__validate_with_womtool(wdl, input_file, imports_file)

        logger.info('cmd: {cmd}'.format(cmd=cmd))
        if self._dry_run:
            return -1
        try:
            p = Popen(cmd, stdout=PIPE, universal_newlines=True)
            workflow_id = None
            rc = None
            while p.poll() is None:
                stdout = p.stdout.readline().strip('\n')

                # find workflow id from STDOUT
                if workflow_id is None:
                    wf_ids_with_status = \
                        Caper.__get_workflow_ids_from_cromwell_stdout(
                            stdout)
                    for wf_id, status in wf_ids_with_status:
                        if status == 'started' or status == 'finished':
                            workflow_id = wf_id
                            break
                if stdout != '':
                    print(stdout)
            # get final RC
            rc = p.poll()
        except CalledProcessError as e:
            rc = e.returncode
        except KeyboardInterrupt:
            logger.error(Caper.USER_INTERRUPT_WARNING)
            while p.poll() is None:
                stdout = p.stdout.readline().strip('\n')
                print(stdout)

        # move metadata file to a workflow output directory
        if metadata_file is not None and workflow_id is not None and \
                os.path.exists(metadata_file):
            with open(metadata_file, 'r') as fp:
                metadata_uri = self.__write_metadata_json(
                    workflow_id,
                    json.loads(fp.read()))
            # remove original one
            os.remove(metadata_file)
        else:
            metadata_uri = None

        # troubleshoot if metadata is available
        if metadata_uri is not None:
            Caper.__troubleshoot(
                AbsPath.localize(metadata_uri),
                self._show_completed_task)

        logger.info(
            'run: {rc}, {wf_id}, {m}'.format(
                rc=rc, wf_id=workflow_id, m=metadata_uri))
        return workflow_id

    def server(self):
        """Run a Cromwell server
        """
        tmp_dir = self.__mkdir_tmp_dir()
        backend_file = self.__create_backend_conf_file(tmp_dir)

        # check if port is open
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((self._ip, self._port))
        if result == 0:
            err = '[Caper] Error: server port {} is already taken. '\
                  'Try with a different --port'.format(self._port)
            raise Exception(err)

        java_heap = '-Xmx{}'.format(self._java_heap_server)
        # LOG_LEVEL must be >=INFO to catch workflow ID from STDOUT
        cmd = ['java', java_heap, '-XX:ParallelGCThreads=1', '-DLOG_LEVEL=INFO',
               '-DLOG_MODE=standard',
               '-jar', '-Dconfig.file={}'.format(backend_file),
               install_cromwell_jar(self._cromwell), 'server']
        logger.info('cmd: {cmd}'.format(cmd=cmd))

        # pending/running workflows
        started_wf_ids = set()
        # completed, aborted or terminated workflows
        finished_wf_ids = set()

        self._stop_heartbeat_thread = False
        t_heartbeat = Thread(
            target=self.__write_heartbeat_file)
        if self._dry_run:
            return -1
        try:
            p = Popen(cmd, stdout=PIPE, universal_newlines=True)
            rc = None
            t0 = time.perf_counter()  # tickcount
            server_is_ready = False

            while p.poll() is None:
                stdout = p.stdout.readline().strip('\n')
                if stdout != '':
                    print(stdout)

                # find workflow id from Cromwell server's STDOUT
                wf_ids_with_status = \
                    Caper.__get_workflow_ids_from_cromwell_stdout(stdout)
                if not server_is_ready:
                    server_is_ready = \
                        Caper.__check_cromwell_server_start_from_stdout(stdout)
                    if server_is_ready:
                        t_heartbeat.start()

                for wf_id, status in wf_ids_with_status:
                    if status in 'started':
                        started_wf_ids.add(wf_id)
                    elif status == 'finished':
                        finished_wf_ids.add(wf_id)

                for wf_id in finished_wf_ids:
                    started_wf_ids.remove(wf_id)

                # write metadata.json for finished workflows
                self.__write_metadata_jsons(finished_wf_ids)
                # flush finished workflow IDs
                #   so that their metadata don't get updated any longer
                finished_wf_ids.clear()

                # write metadata.json for running workflows
                #   every SEC_INTERVAL_UPDATE_METADATA
                t1 = time.perf_counter()
                if (t1 - t0) > Caper.SEC_INTERVAL_UPDATE_METADATA:
                    t0 = t1
                    self.__write_metadata_jsons(started_wf_ids)

        except CalledProcessError as e:
            rc = e.returncode
        except KeyboardInterrupt:
            logger.error(Caper.USER_INTERRUPT_WARNING)
            while p.poll() is None:
                stdout = p.stdout.readline().strip('\n')
                print(stdout)
        time.sleep(1)
        self._stop_heartbeat_thread = True
        t_heartbeat.join()
        logger.info(
            'server: {rc}, {s_wf_ids}, {f_wf_ids}'.format(
                rc=rc, s_wf_ids=started_wf_ids, f_wf_ids=finished_wf_ids))
        return rc

    def submit(self):
        """Submit a workflow to Cromwell server
        """
        timestamp = Caper.__get_time_str()
        # otherwise, use WDL basename
        suffix = os.path.join(
            self.__get_wdl_basename_wo_ext(), timestamp)
        tmp_dir = self.__mkdir_tmp_dir(suffix)

        # all input files
        input_file = self.__create_input_json_file(tmp_dir)
        if self._imports is not None:
            imports_file = AbsPath.localize(self._imports)
        else:
            imports_file = self.__create_imports_zip_file_from_wdl(tmp_dir)
        workflow_opts_file = self.__create_workflow_opts_json_file(
            input_file, tmp_dir)
        labels_file = self.__create_labels_json_file(tmp_dir)
        on_hold = self._hold if self._hold is not None else False
        wdl = AbsPath.localize(self._wdl)

        logger.debug(
            'submit params: wdl={w}, imports_f={imp}, input_f={i}, '
            'opt_f={o}, labels_f={l}, on_hold={on_hold}'.format(
                w=wdl,
                imp=imports_file,
                i=input_file,
                o=workflow_opts_file,
                l=labels_file,
                on_hold=on_hold))

        self.__validate_with_womtool(wdl, input_file, imports_file)

        if self._dry_run:
            return -1
        r = self._cromwell_rest_api.submit(
            source=wdl,
            dependencies=imports_file,
            inputs_file=input_file,
            options_file=workflow_opts_file,
            labels_file=labels_file,
            on_hold=on_hold)
        logger.info('submit: {r}'.format(r=r))
        return r

    def abort(self):
        """Abort running/pending workflows on a Cromwell server
        """
        if self._dry_run:
            return -1
        r = self._cromwell_rest_api.abort(
                self._wf_id_or_label,
                [(Caper.KEY_CAPER_STR_LABEL, v)
                 for v in self._wf_id_or_label])
        logger.info('abort: {r}'.format(r=r))
        return r

    def unhold(self):
        """Release hold of workflows on a Cromwell server
        """
        if self._dry_run:
            return -1
        r = self._cromwell_rest_api.release_hold(
                self._wf_id_or_label,
                [(Caper.KEY_CAPER_STR_LABEL, v)
                 for v in self._wf_id_or_label])
        logger.info('unhold: {r}'.format(r=r))
        return r

    def metadata(self, no_print=False):
        """Retrieve metadata for workflows from a Cromwell server
        """
        if self._dry_run:
            return -1
        m = self._cromwell_rest_api.get_metadata(
                self._wf_id_or_label,
                [(Caper.KEY_CAPER_STR_LABEL, v)
                 for v in self._wf_id_or_label])
        if not no_print:
            if len(m) == 1:
                m_ = m[0]
            else:
                m_ = m
            print(json.dumps(m_, indent=4))
        return m

    def list(self):
        """Get list of running/pending workflows from a Cromwell server
        """
        if self._dry_run:
            return -1
        # if not argument, list all workflows using wildcard (*)
        if self._wf_id_or_label is None or len(self._wf_id_or_label) == 0:
            workflow_ids = ['*']
            labels = [(Caper.KEY_CAPER_STR_LABEL, '*')]
        else:
            workflow_ids = self._wf_id_or_label
            labels = [(Caper.KEY_CAPER_STR_LABEL, v)
                      for v in self._wf_id_or_label]

        workflows = self._cromwell_rest_api.find(
            workflow_ids, labels)
        formats = self._format.split(',')
        print('\t'.join(formats))

        if workflows is None:
            return None
        for w in workflows:
            row = []
            workflow_id = w['id'] if 'id' in w else None

            if self._hide_result_before is not None:
                if 'submission' in w and w['submission'] <= self._hide_result_before:
                    continue
            for f in formats:
                if f == 'workflow_id':
                    row.append(str(workflow_id))
                elif f == 'str_label':
                    if 'labels' in w and Caper.KEY_CAPER_STR_LABEL in w['labels']:
                        lbl = w['labels'][Caper.KEY_CAPER_STR_LABEL]
                    else:
                        lbl = None
                    row.append(str(lbl))
                elif f == 'user':
                    if 'labels' in w and Caper.KEY_CAPER_USER in w['labels']:
                        lbl = w['labels'][Caper.KEY_CAPER_USER]
                    else:
                        lbl = None
                    row.append(str(lbl))
                else:
                    row.append(str(w[f] if f in w else None))
            print('\t'.join(row))
        return workflows

    def troubleshoot(self):
        """Troubleshoot errors based on information from Cromwell's metadata
        """
        if self._dry_run:
            return -1
        if self._wf_id_or_label is None or len(self._wf_id_or_label) == 0:
            return
        # if it's a file
        wf_id_or_label = []
        metadatas = []
        for f in self._wf_id_or_label:
            u = AutoURI(f)
            if u.is_valid and u.exists:
                metadatas.append(AbsPath.localize(u))
            else:
                wf_id_or_label.append(f)

        if len(wf_id_or_label) > 0:
            self._wf_id_or_label = wf_id_or_label
            metadatas.extend(self.metadata(no_print=True))

        for metadata in metadatas:
            Caper.__troubleshoot(metadata, self._show_completed_task)

    def __validate_with_womtool(self, wdl, input_file, imports):
        if not self._ignore_womtool:
            with TemporaryDirectory() as tmp_d:
                if imports:
                    # copy WDL to temp dir and unpack imports.zip (sub WDLs) if exists
                    wdl_copy = os.path.join(tmp_d, AutoURI(self._wdl).basename)
                    AutoURI(wdl).cp(wdl_copy)
                    shutil.unpack_archive(imports, tmp_d)
                else:
                    wdl_copy = wdl
                cmd_womtool = ['java', '-Xmx512M', '-jar', '-DLOG_LEVEL=INFO',
                               install_womtool_jar(self._womtool),
                               'validate', wdl_copy,
                               '-i', input_file]
                try:
                    logger.info('Validating WDL/input JSON with womtool...')
                    check_call(cmd_womtool)
                except CalledProcessError as e:
                    logger.error('Womtool: WDL or input JSON is invalid '
                                 'or input JSON doesn\'t exist.')
                    rc = e.returncode
                    sys.exit(rc)

    def __init_cromwell_rest_api(self, action, ip, port,
                                 no_server_hearbeat,
                                 server_hearbeat_file,
                                 server_hearbeat_timeout):
        self._no_server_hearbeat = no_server_hearbeat
        self._server_hearbeat_file = server_hearbeat_file
        self._ip, self._port = \
            self.__read_heartbeat_file(action, ip, port, server_hearbeat_timeout)

        self._cromwell_rest_api = CromwellRestAPI(
            ip=self._ip, port=self._port, verbose=False)

    def __read_heartbeat_file(self, action, ip, port, server_hearbeat_timeout):
        if not self._no_server_hearbeat and self._server_hearbeat_file is not None:
            self._server_hearbeat_file = os.path.expanduser(
                self._server_hearbeat_file)
            if action != 'server':
                try:
                    if os.path.exists(self._server_hearbeat_file):
                        f_time = os.path.getmtime(self._server_hearbeat_file)
                        if (time.time() - f_time) * 1000.0 < server_hearbeat_timeout:
                            with open(self._server_hearbeat_file, 'r') as fp:
                                ip, port = fp.read().strip('\n').split(':')
                except:
                    logger.warning(
                        'Failed to read server_heartbeat_file: {f}'.format(
                            f=self._server_hearbeat_file))
        return ip, port

    def __write_heartbeat_file(self):
        if not self._no_server_hearbeat and self._server_hearbeat_file is not None:
            while True:
                try:
                    logger.info(
                        'Writing heartbeat: {ip}, {port}'.format(
                            ip=socket.gethostname(), port=self._port))
                    with open(self._server_hearbeat_file, 'w') as fp:
                        fp.write('{}:{}'.format(
                            socket.gethostname(),
                            self._port))
                except Exception as e:
                    logger.warning('Failed to write a heartbeat_file')
                cnt = 0
                while cnt < Caper.SEC_INTERVAL_UPDATE_SERVER_HEARTBEAT:
                    cnt += 1
                    if self._stop_heartbeat_thread:
                        break
                    time.sleep(1)
                if self._stop_heartbeat_thread:
                    break

    def __write_metadata_jsons(self, workflow_ids):
        for wf_id in workflow_ids.copy():
            for trial in range(Caper.MAX_RETRY_UPDATING_METADATA + 1):
                try:
                    time.sleep(Caper.SEC_INTERVAL_RETRY_UPDATING_METADATA)
                    # get metadata for wf_id
                    m = self._cromwell_rest_api.get_metadata([wf_id])
                    assert(len(m) == 1)
                    metadata = m[0]
                    if 'labels' in metadata and \
                            'caper-backend' in metadata['labels']:
                        backend = \
                            metadata['labels']['caper-backend']
                    else:
                        backend = None

                    if backend is not None:
                        self.__write_metadata_json(
                            wf_id, metadata,
                            backend=backend,
                            wdl=metadata['workflowName'])
                except Exception as e:
                    logger.warning(
                        '[Caper] Exception caught while retrieving '
                        'metadata from Cromwell server. '
                        'trial: {t}, wf_id: {wf_id}, e: {e}'.format(
                            t=trial, wf_id=wf_id, e=str(e)))
                break

    def __write_metadata_json(self, workflow_id, metadata_json,
                              backend=None, wdl=None):
        if backend is None:
            backend = self._backend
        if backend is None:
            return None

        if backend == BACKEND_GCP:
            out_dir = self._out_gcs_bucket
        elif backend == BACKEND_AWS:
            out_dir = self._out_s3_bucket
        else:
            out_dir = self._out_dir

        if wdl is None:
            wdl = self.__get_wdl_basename_wo_ext()
        if wdl is None:
            path = os.path.join(out_dir, workflow_id)
        else:
            path = os.path.join(out_dir, os.path.basename(wdl), workflow_id)

        if self._metadata_output is not None:
            metadata_uri = self._metadata_output
        else:
            metadata_uri = os.path.join(
                path, Caper.TMP_FILE_BASENAME_METADATA_JSON)

        return AutoURI(metadata_uri).write(
            json.dumps(metadata_json, indent=4))

    def __create_input_json_file(
            self, directory, fname='inputs.json'):
        """Make a copy of input JSON file.
        Deepcopy to a specified storage if required.
        """
        if self._inputs is not None:
            if not self._no_deepcopy:
                # deepcopy all files in JSON/TSV/CSV
                #   to the target backend
                if self._backend == BACKEND_GCP:
                    uri_cls = GCSURI
                elif self._backend == BACKEND_AWS:
                    uri_cls = S3URI
                else:
                    uri_cls = AbsPath

                new_uri = uri_cls.localize(
                    self._inputs,
                    recursive=True,
                    make_md5_file=True)
            else:
                new_uri = self._inputs
            # localize again on local
            return AbsPath.localize(new_uri)
        else:
            input_file = os.path.join(directory, fname)
            with open(input_file, 'w') as fp:
                fp.write('{}')
            return input_file

    def __create_labels_json_file(
            self, directory, fname=TMP_FILE_BASENAME_LABELS_JSON):
        """Create labels JSON file
        """
        if self._labels is not None:
            s = AutoURI(self._labels).read()
            labels_dict = json.loads(s)
        else:
            labels_dict = {}

        labels_dict[Caper.KEY_CAPER_BACKEND] = self._backend
        if self._str_label is not None:
            labels_dict[Caper.KEY_CAPER_STR_LABEL] = \
                self._str_label
        username = pwd.getpwuid(os.getuid())[0]
        labels_dict[Caper.KEY_CAPER_USER] = username

        labels_file = os.path.join(directory, fname)
        with open(labels_file, 'w') as fp:
            fp.write(json.dumps(labels_dict, indent=4))
        return labels_file

    def __create_workflow_opts_json_file(
            self, input_json_file,
            directory, fname=TMP_FILE_BASENAME_WORKFLOW_OPTS_JSON,):
        """Creates Cromwell's workflow options JSON file

        input_json_file is required to find singularity_bindpath,
        which is a common root for all data input JSON.

        Items written to workflow options JSON file:
            * very important backend
            backend: a backend to run workflows on

            * important dep resolver
            docker:
                docker image URI (e.g. ubuntu:latest)
            singularity:
                singularity image URI (docker://, shub://)
                singularity_bindpath (calculate common root of
                                      files in input JSON)
                singularity_cachedir

            * SLURM params (can also be defined in backend conf file)
            slurm_partition
            slurm_account
            slurm_extra_param

            * SGE params (can also be defined in backend conf file)
            sge_pe
            sge_queue
            sge_extra_param

            * PBS params (can also be defined in backend conf file)
            pbs_queue
            pbs_extra_param
        """
        template = {
            'default_runtime_attributes': {}
        }

        if self._backend is not None:
            template['backend'] = \
                self._backend

        # automatically add docker_from_wdl for cloud backend
        if self._use_docker or self._backend in (BACKEND_GCP, BACKEND_AWS):
            if self._docker is None:
                # find docker from WDL or command line args
                docker = CaperWDLParser(self._wdl).find_docker()
            else:
                docker = self._docker
            if docker is None:
                logger.warning('No docker image specified with --docker.')
            else:
                template['default_runtime_attributes']['docker'] = docker

        if self._use_singularity:
            if self._singularity is None:
                # find singularity from WDL or command line args
                singularity = CaperWDLParser(self._wdl).find_singularity()
            else:
                singularity = self._singularity
            if singularity is None:
                raise ValueError('Singularity image is not defined either '
                                 'in WDL nor in --singularity.')
            # build singularity image before running a pipeline
            self.__build_singularity_image(singularity)

            # important singularity settings (cachedir, bindpath)
            template['default_runtime_attributes']['singularity'] = singularity
            if self._singularity_cachedir is not None:
                template['default_runtime_attributes']['singularity_cachedir'] = \
                        self._singularity_cachedir
            # calculate bindpath from all file paths in input JSON file
            template['default_runtime_attributes']['singularity_bindpath'] = \
                Caper.__find_singularity_bindpath(input_json_file)

        if self._gcp_zones is not None:
            # delimiters: comma or whitespace
            zones = ' '.join(re.split(Caper.RE_PATTERN_DELIMITER_GCP_ZONES,
                                      self._gcp_zones))
            template['default_runtime_attributes']['zones'] = zones

        if self._slurm_partition is not None:
            template['default_runtime_attributes']['slurm_partition'] = \
                self._slurm_partition
        if self._slurm_account is not None:
            template['default_runtime_attributes']['slurm_account'] = \
                self._slurm_account
        if self._slurm_extra_param is not None:
            template['default_runtime_attributes']['slurm_extra_param'] = \
                self._slurm_extra_param

        if self._pbs_queue is not None:
            template['default_runtime_attributes']['pbs_queue'] = \
                self._pbs_queue
        if self._pbs_extra_param is not None:
            template['default_runtime_attributes']['pbs_extra_param'] = \
                self._pbs_extra_param

        if self._sge_pe is not None:
            template['default_runtime_attributes']['sge_pe'] = \
                self._sge_pe
        if self._sge_queue is not None:
            template['default_runtime_attributes']['sge_queue'] = \
                self._sge_queue
        if self._sge_extra_param is not None:
            template['default_runtime_attributes']['sge_extra_param'] = \
                self._sge_extra_param

        if self._max_retries is not None:
            template['default_runtime_attributes']['maxRetries'] = \
                self._max_retries

        # if workflow opts file is given by a user, merge it to template
        if self._workflow_opts is not None:
            f = AbsPath.localize(self._workflow_opts)
            with open(f, 'r') as fp:
                d = json.loads(fp.read())
                merge_dict(template, d)

        # write it
        workflow_opts_file = os.path.join(directory, fname)
        with open(workflow_opts_file, 'w') as fp:
            fp.write(json.dumps(template, indent=4))
            fp.write('\n')

        return workflow_opts_file

    def __create_imports_zip_file_from_wdl(
            self, directory, fname=TMP_FILE_BASENAME_IMPORTS_ZIP):
        zip_file = os.path.join(directory, fname)
        if CaperWDLParser(self._wdl).zip_subworkflows(zip_file):
            return zip_file
        return None

    def __create_backend_conf_file(
            self, directory, fname=TMP_FILE_BASENAME_BACKEND_CONF):
        """Creates Cromwell's backend conf file
        """
        backend_str = self.__get_backend_conf_str()
        backend_file = os.path.join(directory, fname)
        with open(backend_file, 'w') as fp:
            fp.write(backend_str)
        return backend_file

    def __get_backend_conf_str(self):
        """
        Initializes the following backend stanzas,
        which are defined in "backend" {} in a Cromwell's backend
        configuration file:
            1) local: local backend
            2) gc: Google Cloud backend (optional)
            3) aws: AWS backend (optional)
            4) slurm: SLURM (optional)
            5) sge: SGE (optional)
            6) pbs: PBS (optional)

        Also, initializes the following common non-"backend" stanzas:
            a) common: base stanzas
            b) mysql: connect to MySQL (optional)

        Then converts it to a HOCON string
        """
        # init backend dict
        backend_dict = {}

        # common stanza for backend conf file
        merge_dict(
            backend_dict,
            CaperBackendCommon(
                port=self._port,
                disable_call_caching=self._disable_call_caching,
                max_concurrent_workflows=self._max_concurrent_workflows))

        # common settings for all backends
        if self._max_concurrent_tasks is not None:
            CaperBackendBase.CONCURRENT_JOB_LIMIT = self._max_concurrent_tasks

        # common settings for local-based backends
        if self._soft_glob_output is not None:
            CaperBackendBaseLocal.USE_SOFT_GLOB_OUTPUT = self._soft_glob_output
        if self._out_dir is not None:
            CaperBackendBaseLocal.OUT_DIR = self._out_dir

        # local backend
        merge_dict(
            backend_dict,
            CaperBackendLocal())

        # GC
        if self._gcp_prj is not None and self._out_gcs_bucket is not None:
            merge_dict(
                backend_dict,
                CaperBackendGCP(
                    gcp_prj=self._gcp_prj,
                    out_gcs_bucket=self._out_gcs_bucket,
                    call_caching_dup_strat=self._gcp_call_caching_dup_strat))

        # AWS
        if self._aws_batch_arn is not None and self._aws_region is not None \
                and self._out_s3_bucket is not None:
            merge_dict(
                backend_dict,
                CaperBackendAWS(
                    aws_batch_arn=self._aws_batch_arn,
                    aws_region=self._aws_region,
                    out_s3_bucket=self._out_s3_bucket))

        # SLURM
        merge_dict(
            backend_dict,
            CaperBackendSLURM(
                partition=self._slurm_partition,
                account=self._slurm_account,
                extra_param=self._slurm_extra_param))

        # SGE
        merge_dict(
            backend_dict,
            CaperBackendSGE(
                pe=self._sge_pe,
                queue=self._sge_queue,
                extra_param=self._sge_extra_param))

        # PBS
        merge_dict(
            backend_dict,
            CaperBackendPBS(
                queue=self._pbs_queue,
                extra_param=self._pbs_extra_param))

        # Database
        merge_dict(
            backend_dict,
            CaperBackendDatabase(
                db_type=self._db,
                db_timeout=self._db_timeout,
                file_db=self._file_db,
                mysql_ip=self._mysql_db_ip,
                mysql_port=self._mysql_db_port,
                mysql_user=self._mysql_db_user,
                mysql_password=self._mysql_db_password,
                mysql_name=self._mysql_db_name,
                postgresql_ip=self._postgresql_db_ip,
                postgresql_port=self._postgresql_db_port,
                postgresql_user=self._postgresql_db_user,
                postgresql_password=self._postgresql_db_password,
                postgresql_name=self._postgresql_db_name))

        # set header for conf ("include ...")
        assert(Caper.BACKEND_CONF_HEADER.endswith('\n'))
        lines_header = [Caper.BACKEND_CONF_HEADER]

        # override with user-specified backend.conf if exists
        if self._backend_file is not None:
            lines_wo_header = []

            with open(AbsPath.localize(self._backend_file),
                      'r') as fp:
                for line in fp.readlines():
                    # find header and exclude
                    if re.findall(Caper.RE_PATTERN_BACKEND_CONF_HEADER,
                                  line):
                        if line not in lines_header:
                            lines_header.append(line)
                    else:
                        lines_wo_header.append(line)

            # parse HOCON to JSON to dict
            c = ConfigFactory.parse_string(''.join(lines_wo_header))
            j = HOCONConverter.to_json(c)
            d = json.loads(j)
            # apply to backend conf
            merge_dict(backend_dict, d)

        # use default backend (local) if not specified
        if self._backend is not None:
            backend_dict['backend']['default'] = self._backend
        else:
            backend_dict['backend']['default'] = Caper.DEFAULT_BACKEND

        # dict to HOCON (excluding header)
        backend_hocon = ConfigFactory.from_dict(backend_dict)
        # write header to HOCON string
        backend_str = ''.join(lines_header)
        # convert HOCON to string
        backend_str += HOCONConverter.to_hocon(backend_hocon)

        return backend_str

    def __get_wdl_basename_wo_ext(self):
        if self._wdl is not None:
            wdl, _ = os.path.splitext(self._wdl)
            return os.path.basename(wdl)
        else:
            return None

    def __mkdir_tmp_dir(self, suffix=''):
        """Create a temporary directory (self._tmp_dir/suffix)
        """
        tmp_dir = os.path.join(self._tmp_dir, suffix)
        os.makedirs(tmp_dir, exist_ok=True)
        return tmp_dir

    def __build_singularity_image(self, singularity):
        if self._no_build_singularity is not None \
                and self._no_build_singularity:
            pass

        elif self._backend is not None \
                and self._backend not in (BACKEND_AWS, BACKEND_GCP):
            logger.info(
                'Building local singularity image: {img}'.format(
                    img=singularity))
            cmd = ['singularity', 'exec', singularity,
                   'echo', '[Caper] building done.']
            env = os.environ.copy()
            if self._singularity_cachedir is not None \
                    and 'SINGULARITY_CACHEDIR' not in env:
                env['SINGULARITY_CACHEDIR'] = self._singularity_cachedir
            return check_call(cmd, env=env)

        logger.info('Skipped building local singularity image.')
        return None

    @staticmethod
    def __get_workflow_ids_from_cromwell_stdout(stdout):
        result = []
        for line in stdout.split('\n'):
            r1 = re.findall(Caper.RE_PATTERN_STARTED_WORKFLOW_ID, line)
            if len(r1) > 0:
                result.append((r1[0].strip(), 'started'))
            r2 = re.findall(Caper.RE_PATTERN_FINISHED_WORKFLOW_ID, line)
            if len(r2) > 0:
                result.append((r2[0].strip(), 'finished'))
        return result

    @staticmethod
    def __check_cromwell_server_start_from_stdout(stdout):
        for line in stdout.split('\n'):
            r1 = re.findall(Caper.RE_PATTERN_STARTED_CROMWELL_SERVER, line)
            if len(r1) > 0:
                return True
        return False

    @staticmethod
    def __get_time_str():
        return datetime.now().strftime('%Y%m%d_%H%M%S_%f')

    @staticmethod
    def __troubleshoot(metadata_json, show_completed_task=False):
        """Troubleshoot from metadata JSON obj/file
        """
        if isinstance(metadata_json, dict):
            metadata = metadata_json
        else:
            f = AbsPath.localize(metadata_json)
            with open(f, 'r') as fp:
                metadata = json.loads(fp.read())
        if isinstance(metadata, list):
            metadata = metadata[0]

        workflow_id = metadata['id']
        workflow_status = metadata['status']
        logger.info('Troubleshooting {wf_id} ...'.format(wf_id=workflow_id))
        if not show_completed_task and workflow_status == 'Succeeded':
            logger.info(
                'This workflow ran successfully. There is nothing to troubleshoot')
            return

        def recurse_calls(calls, failures=None, show_completed_task=False):
            if failures is not None:
                s = json.dumps(failures, indent=4)
                logger.info('Found failures:\n{s}'.format(s=s))
            for task_name, call_ in calls.items():
                for call in call_:
                    # if it is a subworkflow, then recursively dive into it
                    if 'subWorkflowMetadata' in call:
                        subworkflow = call['subWorkflowMetadata']
                        recurse_calls(
                            subworkflow['calls'],
                            subworkflow['failures']
                            if 'failures' in subworkflow else None,
                            show_completed_task)
                        continue
                    task_status = call['executionStatus']
                    shard_index = call['shardIndex']
                    rc = call['returnCode'] if 'returnCode' in call else None
                    job_id = call['jobId'] if 'jobId' in call else None
                    stdout = call['stdout'] if 'stdout' in call else None
                    stderr = call['stderr'] if 'stderr' in call else None
                    run_start = None
                    run_end = None
                    if 'executionEvents' in call:
                        for ev in call['executionEvents']:
                            if ev['description'].startswith('Running'):
                                run_start = ev['startTime']
                                run_end = ev['endTime']
                                break

                    if not show_completed_task and \
                            task_status in ('Done', 'Succeeded'):
                        continue
                    print(
                        '\n{tn} {ts}. SHARD_IDX={shard_id}, RC={rc}, JOB_ID={job_id}, '
                        'RUN_START={start}, RUN_END={end}, '
                        'STDOUT={stdout}, STDERR={stderr}'.format(
                            tn=task_name, ts=task_status,
                            shard_id=shard_index, rc=rc, job_id=job_id,
                            start=run_start, end=run_end,
                            stdout=stdout, stderr=stderr))

                    if stderr is not None:
                        u = AutoURI(stderr)
                        if u.is_valid and u.exists:
                            local_stderr_f = AbsPath.localize(u)
                            with open(local_stderr_f, 'r') as fp:
                                stderr_contents = fp.read()
                            print('STDERR_CONTENTS=\n{}'.format(
                                stderr_contents))

        calls = metadata['calls']
        failures = metadata['failures'] if 'failures' in metadata else None
        recurse_calls(calls, failures, show_completed_task)

    @staticmethod
    def __find_singularity_bindpath(input_json_file):
        """Find paths to be bound for singularity
        by finding common roots for all files in input JSON file.
        This function will recursively visit all values in input JSON and
        also JSON, TSV, CSV files in the input JSON itself.

        This function visit all files in input JSON.
        Files with some extensions (defined by Autouri's URIBase.LOC_RECURSE_EXT_AND_FNC)
        are recursively visited.

        Args:
            input_json_file:
                localized input JSON file so that all files in it are already
                recursively localized
        """
        with open(input_json_file, 'r') as fp:
            input_json_contents = fp.read()

        all_dirnames = []
        def find_dirname(s):
            u = AbsPath(s)
            if u.is_valid:
                for ext, recurse_fnc_for_ext in URIBase.LOC_RECURSE_EXT_AND_FNC.items():
                    if u.ext == ext:
                        _, _ = recurse_fnc_for_ext(u.read(), find_dirname)
                # file can be a soft-link
                # singularity will want to have access to both soft-link and real one
                # so add dirnames of both soft-link and realpath
                all_dirnames.append(u.dirname)
                all_dirnames.append(os.path.dirname(os.path.realpath(u.uri)))
            return None, False
        _, _ = recurse_json(input_json_contents, find_dirname)

        # add all (but not too high level<4) parent directories
        # to all_dirnames. start from original
        # e.g. /a/b/c/d/e/f/g/h with COMMON_ROOT_SEARCH_LEVEL = 5
        # add all the followings:
        # /a/b/c/d/e/f/g/h (org)
        # /a/b/c/d/e/f/g
        # /a/b/c/d/e/f
        # /a/b/c/d/e
        # /a/b/c/d (minimum level = COMMON_ROOT_SEARCH_LEVEL-1)
        all_dnames_incl_parents = set(all_dirnames)
        for d in all_dirnames:
            dir_arr = d.split(os.sep)
            for i, _ in enumerate(
                    dir_arr[Caper.COMMON_ROOT_SEARCH_LEVEL:]):
                d_child = os.sep.join(
                    dir_arr[:i + Caper.COMMON_ROOT_SEARCH_LEVEL])
                all_dnames_incl_parents.add(d_child)

        bindpaths = set()
        # remove overlapping directories
        for i, d1 in enumerate(sorted(all_dnames_incl_parents,
                                      reverse=True)):
            overlap_found = False
            for j, d2 in enumerate(sorted(all_dnames_incl_parents,
                                          reverse=True)):
                if i >= j:
                    continue
                if d1.startswith(d2):
                    overlap_found = True
                    break
            if not overlap_found:
                bindpaths.add(d1)

        return ','.join(bindpaths)


def main():
    # parse arguments
    #   note that args is a dict
    args = parse_caper_arguments()
    action = args['action']
    if action == 'init':
        init_caper_conf(args)
        sys.exit(0)
    args = check_caper_conf(args)

    # init Autouri classes to transfer files across various storages
    #   e.g. gs:// to s3://, http:// to local, ...
    # loc_prefix means prefix (root directory)
    # for localizing files of different storages
    AbsPath.init_abspath(
        loc_prefix=args.get('tmp_dir')
    )
    GCSURI.init_gcsuri(
        loc_prefix=args.get('tmp_gcs_bucket'),
        use_gsutil_for_s3=args.get('use_gsutil_for_s3')
    )
    S3URI.init_s3uri(
        loc_prefix=args.get('tmp_s3_bucket')
    )
    # init both loggers of Autouri and Caper
    if args.get('verbose'):
        autouri_logger.setLevel('INFO')
        logger.setLevel('INFO')
    elif args.get('debug'):
        autouri_logger.setLevel('DEBUG')
        logger.setLevel('DEBUG')

    # initialize caper: taking all args at init step
    c = Caper(args)

    if action == 'run':
        c.run()
    elif action == 'server':
        c.server()
    elif action == 'submit':
        c.submit()
    elif action == 'abort':
        c.abort()
    elif action == 'list':
        c.list()
    elif action == 'metadata':
        c.metadata()
    elif action == 'unhold':
        c.unhold()
    elif action in ['troubleshoot', 'debug']:
        c.troubleshoot()

    else:
        raise Exception('Unsupported or unspecified action.')
    return 0


if __name__ == '__main__':
    main()



"""CromwellRestAPI
"""

import requests
import io
import fnmatch
import json
import sys
# import traceback


class CromwellRestAPI(object):
    QUERY_URL = 'http://{ip}:{port}'
    ENDPOINT_BACKEND = '/api/workflows/v1/backends'
    ENDPOINT_WORKFLOWS = '/api/workflows/v1/query'
    ENDPOINT_METADATA = '/api/workflows/v1/{wf_id}/metadata'
    ENDPOINT_LABELS = '/api/workflows/v1/{wf_id}/labels'
    ENDPOINT_SUBMIT = '/api/workflows/v1'
    ENDPOINT_ABORT = '/api/workflows/v1/{wf_id}/abort'
    ENDPOINT_RELEASE_HOLD = '/api/workflows/v1/{wf_id}/releaseHold'
    KEY_LABEL = 'cromwell_rest_api_label'
    PARAMS_WORKFLOWS = {
        'additionalQueryResultFields': 'labels'
    }

    def __init__(self, ip='localhost', port=8000,
                 user=None, password=None, verbose=False):
        self._verbose = verbose
        self._ip = ip
        self._port = port

        self._user = user
        self._password = password
        self.__init_auth()

    def submit(self, source, dependencies=None, inputs_file=None,
               options_file=None, labels_file=None, on_hold=False):
        """Submit a workflow.

        Returns:
            JSON Response from POST request submit a workflow
        """
        manifest = {}
        manifest['workflowSource'] = \
            CromwellRestAPI.__get_string_io_from_file(source)
        if dependencies is not None:
            manifest['workflowDependencies'] = \
                CromwellRestAPI.__get_bytes_io_from_file(dependencies)
        if inputs_file is not None:
            manifest['workflowInputs'] = \
                CromwellRestAPI.__get_string_io_from_file(inputs_file)
        else:
            manifest['workflowInputs'] = io.StringIO('{}')
        if options_file is not None:
            manifest['workflowOptions'] = \
                CromwellRestAPI.__get_string_io_from_file(options_file)
        if labels_file is not None:
            manifest['labels'] = \
                CromwellRestAPI.__get_string_io_from_file(labels_file)
        if on_hold:
            manifest['workflowOnHold'] = True

        r = self.__request_post(CromwellRestAPI.ENDPOINT_SUBMIT, manifest)
        if self._verbose:
            print("CromwellRestAPI.submit: ", r)
        return r

    def abort(self, workflow_ids=None, labels=None):
        """Abort workflows matching workflow IDs or labels

        Returns:
            List of JSON responses from POST request
            for aborting workflows
        """
        workflows = self.find(workflow_ids, labels)
        if workflows is None:
            return None
        result = []
        for w in workflows:
            r = self.__request_post(
                CromwellRestAPI.ENDPOINT_ABORT.format(
                    wf_id=w['id']))
            result.append(r)
        if self._verbose:
            print("CromwellRestAPI.abort: ", result)
        return result

    def release_hold(self, workflow_ids=None, labels=None):
        """Release hold of workflows matching workflow IDs or labels

        Returns:
            List of JSON responses from POST request
            for releasing hold of workflows
        """
        workflows = self.find(workflow_ids, labels)
        if workflows is None:
            return None
        result = []
        for w in workflows:
            r = self.__request_post(
                CromwellRestAPI.ENDPOINT_RELEASE_HOLD.format(
                    wf_id=w['id']))
            result.append(r)
        if self._verbose:
            print("CromwellRestAPI.release_hold: ", result)
        return result

    def get_default_backend(self):
        """Retrieve default backend name

        Returns:
            Default backend name
        """
        return self.get_backends()['defaultBackend']

    def get_backends(self):
        """Retrieve available backend names and default backend name

        Returns:
            JSON response with keys "defaultBackend" and "supportedBackends"
            Example: {"defaultBackend":"Local","supportedBackends":
                      ["Local","aws","gcp","pbs","sge","slurm"]}
        """
        return self.__request_get(CromwellRestAPI.ENDPOINT_BACKEND)

    def get_metadata(self, workflow_ids=None, labels=None):
        """Retrieve metadata for workflows matching workflow IDs or labels

        Returns:
            List of metadata JSONs
        """
        workflows = self.find(workflow_ids, labels)
        if workflows is None:
            return None
        result = []
        for w in workflows:
            m = self.__request_get(
                CromwellRestAPI.ENDPOINT_METADATA.format(wf_id=w['id']))
            result.append(m)
        if self._verbose:
            print(json.dumps(result, indent=4))
        return result

    def get_labels(self, workflow_id):
        """Get labels JSON for a specified workflow

        Returns:
            Labels JSON for a workflow
        """
        if workflow_id is None:
            return None
        r = self.__request_get(
            CromwellRestAPI.ENDPOINT_LABELS.format(
                wf_id=workflow_id))
        if r is None:
            return None
        return r['labels']

    def get_label(self, workflow_id, key):
        """Get a label for a key in a specified workflow

        Returns:
            Value for a specified key in labels JSON for a workflow
        """
        labels = self.get_labels(workflow_id)
        if labels is None:
            return None
        if key in labels:
            return labels[key]
        else:
            return None

    def update_labels(self, workflow_id, labels):
        """Update labels for a specified workflow with
        a list of (key, val) tuples
        """
        if workflow_id is None or labels is None:
            return None
        r = self.__request_patch(
            CromwellRestAPI.ENDPOINT_LABELS.format(
                wf_id=workflow_id), labels)
        if self._verbose:
            print("CromwellRestAPI.update_labels: ", r)
        return r

    def find(self, workflow_ids=None, labels=None):
        """Find workflows by matching workflow IDs, label (key, value) tuples.
        Wildcards (? and *) are allowed for string workflow IDs and values in
        a tuple label. Search criterion is (workflow_ids OR labels).

        Args:
            workflow_ids:
                List of workflows ID strings: [wf_id, ...].
                OR search for multiple workflow IDs
            labels:
                List of (key, val) tuples: [(key: val), (key2: val2), ...].
                OR search for multiple tuples

        Returns:
            List of matched workflow JSONs
        """
        r = self.__request_get(
            CromwellRestAPI.ENDPOINT_WORKFLOWS,
            params=CromwellRestAPI.PARAMS_WORKFLOWS)
        if r is None:
            return None
        workflows = r['results']
        if workflows is None:
            return None
        matched = set()
        for w in workflows:
            if 'id' not in w:
                continue
            if workflow_ids is not None:
                for wf_id in workflow_ids:
                    if fnmatch.fnmatchcase(w['id'], wf_id):
                        matched.add(w['id'])
                        break
            if w['id'] in matched:
                continue
            if labels is not None and 'labels' in w:
                labels_ = w['labels']
                for k, v in labels:
                    if k in labels_:
                        v_ = labels_[k]
                        if isinstance(v_, str) and isinstance(v, str):
                            # wildcard allowed for str values
                            if fnmatch.fnmatchcase(v_, v):
                                matched.add(w['id'])
                                break
                        elif v_ == v:
                            matched.add(w['id'])
                            break
        result = []
        for w in workflows:
            if 'id' not in w:
                continue
            if w['id'] in matched:
                result.append(w)
        if self._verbose:
            print('CromwellRestAPI.find: ', result)
        return result

    def __init_auth(self):
        """Init auth object
        """
        if self._user is not None and self._password is not None:
            self._auth = (self._user, self._password)
        else:
            self._auth = None

    def __request_get(self, endpoint, params=None):
        """GET request

        Returns:
            JSON response
        """
        url = CromwellRestAPI.QUERY_URL.format(
                ip=self._ip,
                port=self._port) + endpoint
        try:
            resp = requests.get(
                url, auth=self._auth, params=params,
                headers={'accept': 'application/json'})
        except Exception as e:
            # traceback.print_exc()
            print(e)
            print('Help: cannot connect to server. '\
                  'Check if server is dead or still spinning up.')
            sys.exit(1)

        if resp.ok:
            return resp.json()
        else:
            print("HTTP GET error: ", resp.status_code, resp.content,
                  url)
            return None

    def __request_post(self, endpoint, manifest=None):
        """POST request

        Returns:
            JSON response
        """
        url = CromwellRestAPI.QUERY_URL.format(
                ip=self._ip,
                port=self._port) + endpoint
        try:
            resp = requests.post(
                url, files=manifest, auth=self._auth,
                headers={'accept': 'application/json'})
        except Exception as e:
            # traceback.print_exc()
            print(e)
            sys.exit(1)

        if resp.ok:
            return resp.json()
        else:
            print("HTTP POST error: ", resp.status_code, resp.content,
                  url, manifest)
            return None

    def __request_patch(self, endpoint, data):
        """POST request

        Returns:
            JSON response
        """
        url = CromwellRestAPI.QUERY_URL.format(
                ip=self._ip,
                port=self._port) + endpoint
        try:
            resp = requests.patch(
                url, data=data, auth=self._auth,
                headers={'accept': 'application/json',
                         'content-type': 'application/json'})
        except Exception as e:
            # traceback.print_exc()
            print(e)
            sys.exit(1)

        if resp.ok:
            return resp.json()
        else:
            print("HTTP PATCH error: ", resp.status_code, resp.content,
                  url, json)
            return None

    @staticmethod
    def __get_string_io_from_file(fname):
        with open(fname, 'r') as fp:
            return io.StringIO(fp.read())

    def __get_bytes_io_from_file(fname):
        with open(fname, 'rb') as fp:
            return io.BytesIO(fp.read())


def main():
    pass


if __name__ == '__main__':
    main()



"""Caper backend
"""
from collections import UserDict
from copy import deepcopy
from .dict_tool import merge_dict

BACKEND_GCP = 'gcp'
BACKEND_AWS = 'aws'
BACKEND_LOCAL = 'Local'  # should be CAPITAL L
BACKEND_SLURM = 'slurm'
BACKEND_SGE = 'sge'
BACKEND_PBS = 'pbs'
BACKENDS = [BACKEND_GCP, BACKEND_AWS, BACKEND_LOCAL, BACKEND_SLURM,
    BACKEND_SGE, BACKEND_PBS]

BACKEND_ALIAS_LOCAL = 'local'  # with small L
BACKEND_ALIAS_GOOGLE = 'google'
BACKEND_ALIAS_AMAZON = 'amazon'
BACKEND_ALIAS_SHERLOCK = 'sherlock'
BACKEND_ALIAS_SCG = 'scg'
BACKEND_ALIASES = [BACKEND_ALIAS_LOCAL, BACKEND_ALIAS_GOOGLE, BACKEND_ALIAS_AMAZON, BACKEND_ALIAS_SHERLOCK,
    BACKEND_ALIAS_SCG]

BACKENDS_WITH_ALIASES = BACKENDS + BACKEND_ALIASES


def get_backend(backend):
    if backend == BACKEND_ALIAS_GOOGLE:
        backend = BACKEND_GCP
    elif backend == BACKEND_ALIAS_AMAZON:
        backend = BACKEND_AWS
    elif backend == BACKEND_ALIAS_SHERLOCK:
        backend = BACKEND_SLURM
    elif backend == BACKEND_ALIAS_SCG:
        backend = BACKEND_SLURM
    elif backend == BACKEND_ALIAS_LOCAL:
        backend = BACKEND_LOCAL
    if backend not in BACKENDS:
        raise ValueError('Unsupported backend: {}'.format(backend))
    return backend


class CaperBackendCommon(UserDict):
    """Common stanzas for all Caper backends
    """
    TEMPLATE = {
        "backend": {
            "default": BACKEND_LOCAL
        },
        "webservice": {
            "port": 8000
        },
        "services": {
            "LoadController": {
                "class": "cromwell.services.loadcontroller.impl"
                ".LoadControllerServiceActor",
                "config": {
                    # due to issues on stanford sherlock/scg
                    "control-frequency": "21474834 seconds"
                }
            }
        },
        "system": {
            "job-rate-control": {
                "jobs": 1,
                "per": "2 seconds"
            },
            "abort-jobs-on-terminate": True,
            "graceful-server-shutdown": True,
            "max-concurrent-workflows": 40
        },
        "call-caching": {
            "enabled": True,
            "invalidate-bad-cache-results": True
        }
    }

    def __init__(self, port=None, disable_call_caching=None,
                 max_concurrent_workflows=None):
        super().__init__(CaperBackendCommon.TEMPLATE)
        if port is not None:
            self['webservice']['port'] = port
        if disable_call_caching is not None:
            self['call-caching']['enabled'] = not disable_call_caching
        if max_concurrent_workflows is not None:
            self['system']['max-concurrent-workflows'] = \
                max_concurrent_workflows


class CaperBackendDatabase(UserDict):
    """Common stanzas for database
    """
    DB_TYPE_IN_MEMORY = 'in-memory'
    DB_TYPE_FILE = 'file'
    DB_TYPE_MYSQL = 'mysql'
    DB_TYPE_POSTGRESQL = 'postgresql'

    TEMPLATE = {
    }

    TEMPLATE_DB_IN_MEMORY = {
    }

    TEMPLATE_DB_FILE = {
        "db": {
            "url": "jdbc:hsqldb:file:{file};shutdown=false;"
                "hsqldb.tx=mvcc;hsqldb.lob_compressed=true",
            "connectionTimeout": 5000
        }
    }

    TEMPLATE_DB_MYSQL = {
        "profile": "slick.jdbc.MySQLProfile$",
        "db": {
            "driver": "com.mysql.cj.jdbc.Driver",
            "url": "jdbc:mysql://{ip}:{port}/{name}?"
                "allowPublicKeyRetrieval=true&useSSL=false&"
                "rewriteBatchedStatements=true&serverTimezone=UTC",
            "user": "cromwell",
            "password": "cromwell",
            "connectionTimeout": 5000
        }
    }

    TEMPLATE_DB_POSTGRESQL = {
        "profile": "slick.jdbc.PostgresProfile$",
        "db": {
            "driver": "org.postgresql.Driver",
            "url": "jdbc:postgresql://{ip}:{port}/{name}",
            "port": 5432,
            "user": "cromwell",
            "password": "cromwell",
            "connectionTimeout": 5000
        }
    }

    def __init__(self, db_type=None, db_timeout=None,
                 file_db=None,
                 mysql_ip=None, mysql_port=None,
                 mysql_user=None, mysql_password=None,
                 mysql_name=None,
                 postgresql_ip=None, postgresql_port=None,
                 postgresql_user=None, postgresql_password=None,
                 postgresql_name=None):
        super().__init__(CaperBackendDatabase.TEMPLATE)

        if db_type == CaperBackendDatabase.DB_TYPE_IN_MEMORY:
            self['database'] = CaperBackendDatabase.TEMPLATE_DB_IN_MEMORY

        elif db_type == CaperBackendDatabase.DB_TYPE_FILE:
            self['database'] = CaperBackendDatabase.TEMPLATE_DB_FILE
            db = self['database']['db']
            db['url'] = db['url'].format(
                file=file_db)

        elif db_type == CaperBackendDatabase.DB_TYPE_MYSQL:
            self['database'] = CaperBackendDatabase.TEMPLATE_DB_MYSQL            
            db = self['database']['db']
            db['url'] = db['url'].format(
                ip=mysql_ip, port=mysql_port, name=mysql_name)
            db['user'] = mysql_user
            db['password'] = mysql_password

        elif db_type == CaperBackendDatabase.DB_TYPE_POSTGRESQL:
            self['database'] = CaperBackendDatabase.TEMPLATE_DB_POSTGRESQL
            db = self['database']['db']
            db['url'] = db['url'].format(
                ip=postgresql_ip, port=postgresql_port, name=postgresql_name)
            db['port'] = postgresql_port
            db['user'] = postgresql_user
            db['password'] = postgresql_password

        else:
            raise ValueError('Unsupported DB type {}'.format(db_type))

        if db_timeout is not None and 'db' in self['database']:
            self['database']['db']['connectionTimeout'] = db_timeout


class CaperBackendBase(UserDict):
    """Base skeleton backend for all backends
    """
    CONCURRENT_JOB_LIMIT = None
    TEMPLATE = {
        "backend": {
            "providers": {
            }
        }
    }
    TEMPLATE_BACKEND = {
        "actor-factory": None,
        "config": {
            "default-runtime-attributes": {
            },
            "concurrent-job-limit": None
        }
    }

    def __init__(self, dict_to_override_self=None, backend_name=None):
        """
        Args:
            dict_to_override_self: dict to override self
            backend_name: backend name
        """
        if backend_name is None:
            raise ValueError('backend_name must be provided.')
        self._backend_name = backend_name

        super().__init__(deepcopy(CaperBackendBase.TEMPLATE))

        config = self.get_backend_config()

        if CaperBackendBase.CONCURRENT_JOB_LIMIT is None:
            raise ValueError('You must define CaperBackendBase.CONCURRENT_JOB_LIMIT.')
        config['concurrent-job-limit'] = CaperBackendBase.CONCURRENT_JOB_LIMIT

        if dict_to_override_self is not None:
            merge_dict(self, deepcopy(dict_to_override_self))

    @property
    def backend_name(self):
        return self._backend_name

    def get_backend(self):
        if self._backend_name not in self['backend']['providers']:
            self['backend']['providers'][self._backend_name] = \
                    deepcopy(CaperBackendBase.TEMPLATE_BACKEND)
        return self['backend']['providers'][self._backend_name]

    def get_backend_config(self):
        return self.get_backend()['config']


class CaperBackendBaseLocal(CaperBackendBase):
    """Base backend for all local backends (including HPCs with cluster engine)
    """
    USE_SOFT_GLOB_OUTPUT = None
    OUT_DIR = None

    SOFT_GLOB_OUTPUT_CMD = 'ln -sL GLOB_PATTERN GLOB_DIRECTORY 2> /dev/null'

    TEMPLATE_BACKEND = {
        "actor-factory": "cromwell.backend.impl.sfs.config.ConfigBackendLifecycleActorFactory",
        "config": {
            "script-epilogue": "sleep 10 && sync",
            "root": None
        }
    }
    def __init__(self, dict_to_override_self=None, backend_name=None):
        super().__init__(backend_name=backend_name)

        merge_dict(
            self.get_backend(),
            deepcopy(CaperBackendBaseLocal.TEMPLATE_BACKEND))
        config = self.get_backend_config()

        if CaperBackendBaseLocal.USE_SOFT_GLOB_OUTPUT is None:
            raise ValueError('You must define CaperBackendBase.USE_SOFT_GLOB_OUTPUT.')
        if CaperBackendBaseLocal.USE_SOFT_GLOB_OUTPUT:
            config['glob-link-command'] = CaperBackendBaseLocal.SOFT_GLOB_OUTPUT_CMD

        if CaperBackendBaseLocal.OUT_DIR is None:
            raise ValueError('You must define CaperBackendBase.OUT_DIR.')
        config['root'] = CaperBackendBaseLocal.OUT_DIR

        if dict_to_override_self is not None:
            merge_dict(self, deepcopy(dict_to_override_self))


class CaperBackendGCP(CaperBackendBase):
    """Google Cloud backend
    """
    CALL_CACHING_DUP_STRAT_REFERENCE = 'reference'
    CALL_CACHING_DUP_STRAT_COPY = 'copy'

    TEMPLATE = {
        "backend": {
            "providers": {
                BACKEND_GCP: {
                    "actor-factory": "cromwell.backend.google.pipelines.v2alpha1.PipelinesApiLifecycleActorFactory",
                    "config": {
                        "default-runtime-attributes": {
                        },
                        "project": None,
                        "root": None,
                        "genomics-api-queries-per-100-seconds": 1000,
                        "maximum-polling-interval": 600,
                        "genomics": {
                            "auth": "application-default",
                            "compute-service-account": "default",
                            "endpoint-url": "https://genomics.googleapis.com/",
                            "restrict-metadata-access": False
                        },
                        "filesystems": {
                            "gcs": {
                                "auth": "application-default",
                                "caching": {
                                    "duplication-strategy": None
                                }
                            }
                        }
                    }
                }
            }
        },
        "google": {
            "application-name": "cromwell",
            "auths": [
                {
                    "name": "application-default",
                    "scheme": "application_default"
                }
            ]
        }
    }

    def __init__(self, gcp_prj, out_gcs_bucket,
                 call_caching_dup_strat=None):
        super().__init__(
            CaperBackendGCP.TEMPLATE,
            backend_name=BACKEND_GCP)
        config = self.get_backend_config()

        config['project'] = gcp_prj
        config['root'] = out_gcs_bucket

        if not out_gcs_bucket.startswith('gs://'):
            raise ValueError('Wrong GCS bucket URI for out_gcs_bucket: {v}'.format(
                v=out_gcs_bucket))

        if call_caching_dup_strat is None:
            dup_strat = CaperBackendGCP.CALL_CACHING_DUP_STRAT_REFERENCE
        else:
            dup_strat = call_caching_dup_strat
        if dup_strat not in (
            CaperBackendGCP.CALL_CACHING_DUP_STRAT_REFERENCE,
            CaperBackendGCP.CALL_CACHING_DUP_STRAT_COPY):
            raise ValueError('Wrong call_caching_dup_strat: {v}'.format(
                v=call_caching_dup_strat))
        config['filesystems']['gcs']['caching']['duplication-strategy'] = dup_strat


class CaperBackendAWS(CaperBackendBase):
    """AWS backend
    """
    TEMPLATE = {
        "backend": {
            "providers": {
                BACKEND_AWS: {
                    "actor-factory": "cromwell.backend.impl.aws.AwsBatchBackendLifecycleActorFactory",
                    "config": {
                        "default-runtime-attributes": {
                            "queueArn": None
                        },
                        "numSubmitAttempts": 6,
                        "numCreateDefinitionAttempts": 6,
                        "root": None,
                        "auth": "default",
                        "filesystems": {
                            "s3": {
                                "auth": "default"
                            }
                        }
                    }
                },
            }
        },
        "aws": {
            "application-name": "cromwell",
            "auths": [
                {
                    "name": "default",
                    "scheme": "default"
                }
            ],
            "region": None
        },
        "engine": {
            "filesystems": {
                "s3": {
                    "auth": "default"
                }
            }
        }
    }

    def __init__(self, aws_batch_arn, aws_region, out_s3_bucket):
        super().__init__(
            CaperBackendAWS.TEMPLATE,
            backend_name=BACKEND_AWS)
        self[BACKEND_AWS]['region'] = aws_region
        config = self.get_backend_config()
        config['default-runtime-attributes']['queueArn'] = aws_batch_arn
        config['root'] = out_s3_bucket
        if not out_s3_bucket.startswith('s3://'):
            raise ValueError('Wrong S3 bucket URI for out_s3_bucket: {v}'.format(
                v=out_s3_bucket))


class CaperBackendLocal(CaperBackendBaseLocal):
    """Local backend
    """
    RUNTIME_ATTRIBUTES = """
        Int? gpu
        String? docker
        String? docker_user
        String? singularity
        String? singularity_bindpath
        String? singularity_cachedir
    """
    SUBMIT = """
        ${if defined(singularity) then "" else "/bin/bash ${script} #"} \
        if [ -z "$SINGULARITY_BINDPATH" ]; then \
        export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \
        if [ -z "$SINGULARITY_CACHEDIR" ]; then \
        export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi; \
        singularity exec --cleanenv --home ${cwd} \
        ${if defined(gpu) then '--nv' else ''} \
        ${singularity} /bin/bash ${script}
    """
    SUBMIT_DOCKER = """
        # make sure there is no preexisting Docker CID file
        rm -f ${docker_cid}
        # run as in the original configuration without --rm flag (will remove later)
        docker run \
          --cidfile ${docker_cid} \
          -i \
          ${"--user " + docker_user} \
          --entrypoint ${job_shell} \
          -v ${cwd}:${docker_cwd} \
          ${docker} ${docker_script}
    """
    TEMPLATE = {
        "backend": {
            "providers": {
                BACKEND_LOCAL: {
                    "config": {
                        "run-in-background": True,
                        "runtime-attributes": RUNTIME_ATTRIBUTES,
                        "submit": SUBMIT,
                        "submit-docker" : SUBMIT_DOCKER
                    }
                }
            }
        }
    }

    def __init__(self):
        super().__init__(
            CaperBackendLocal.TEMPLATE,
            backend_name=BACKEND_LOCAL)


class CaperBackendSLURM(CaperBackendBaseLocal):
    """SLURM backend
    """
    RUNTIME_ATTRIBUTES = """
        String? docker
        String? docker_user
        Int cpu = 1
        Int? gpu
        Int? time
        Int? memory_mb
        String? slurm_partition
        String? slurm_account
        String? slurm_extra_param
        String? singularity
        String? singularity_bindpath
        String? singularity_cachedir
    """
    # try sbatching up to 3 times every 30 second
    # some busy SLURM clusters spit out error, which results in a failure of the whole workflow
    SUBMIT = """ITER=0; until [ $ITER -ge 3 ]; do
        sbatch \
        --export=ALL \
        -J ${job_name} \
        -D ${cwd} \
        -o ${out} \
        -e ${err} \
        ${"-t " + time*60} \
        -n 1 \
        --ntasks-per-node=1 \
        ${true="--cpus-per-task=" false="" defined(cpu)}${cpu} \
        ${true="--mem=" false="" defined(memory_mb)}${memory_mb} \
        ${"-p " + slurm_partition} \
        ${"--account " + slurm_account} \
        ${true="--gres gpu:" false="" defined(gpu)}${gpu} \
        ${slurm_extra_param} \
        --wrap "${if defined(singularity) then '' else \
            '/bin/bash ${script} #'} \
            if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then \
            export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \
            if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then \
            export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi; \
            singularity exec --cleanenv --home ${cwd} \
            ${if defined(gpu) then '--nv' else ''} \
            ${singularity} /bin/bash ${script}" && break
    ITER=$[$ITER+1]; sleep 30; done
    """
    # squeue every 30 second (up to 3 times)
    # unlike qstat -j JOB_ID, squeue -j JOB_ID doesn't return 1 when there is no such job
    # so we need to use squeue -j JOB_ID --noheader and check if output is empty
    # try polling up to 3 times since squeue fails on some busy SLURM clusters
    # e.g. on Stanford Sherlock, squeue didn't work when server is busy
    CHECK_ALIVE = """for ITER in 1 2 3; do CHK_ALIVE=$(squeue --noheader -j ${job_id} --format=%i | grep ${job_id}); if [ -z "$CHK_ALIVE" ]; then if [ "$ITER" == 3 ]; then /bin/bash -c 'exit 1'; else sleep 30; fi; else echo $CHK_ALIVE; break; fi; done"""
    TEMPLATE = {
        "backend": {
            "providers": {
                BACKEND_SLURM: {
                    "config": {
                        "default-runtime-attributes": {
                            "time": 24
                        },
                        "runtime-attributes": RUNTIME_ATTRIBUTES,
                        "submit": SUBMIT,
                        "kill": "scancel ${job_id}",
                        "exit-code-timeout-seconds": 360,
                        "check-alive": CHECK_ALIVE,
                        "job-id-regex": "Submitted batch job (\\d+).*"
                    }
                }
            }
        }
    }

    def __init__(self, partition=None, account=None, extra_param=None):
        super().__init__(
            CaperBackendSLURM.TEMPLATE,
            backend_name=BACKEND_SLURM)
        config = self.get_backend_config()

        if partition is not None and partition != '':
            config['default-runtime-attributes']['slurm_partition'] = partition
        if account is not None and account != '':
            config['default-runtime-attributes']['slurm_account'] = account
        if extra_param is not None and extra_param != '':
            config['default-runtime-attributes']['slurm_extra_param'] = extra_param


class CaperBackendSGE(CaperBackendBaseLocal):
    """SGE backend
    """
    RUNTIME_ATTRIBUTES = """
        String? docker
        String? docker_user
        String sge_pe = "shm"
        Int cpu = 1
        Int? gpu
        Int? time
        Int? memory_mb
        String? sge_queue
        String? sge_extra_param
        String? singularity
        String? singularity_bindpath
        String? singularity_cachedir
    """
    SUBMIT = """
        echo "${if defined(singularity) then '' else '/bin/bash ${script} #'} \
        if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then \
        export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \
        if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then \
        export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi; \
        singularity exec --cleanenv --home ${cwd} \
        ${if defined(gpu) then '--nv' else ''} \
        ${singularity} /bin/bash ${script}" | qsub \
        -S /bin/sh \
        -terse \
        -b n \
        -N ${job_name} \
        -wd ${cwd} \
        -o ${out} \
        -e ${err} \
        ${if cpu>1 then "-pe " + sge_pe + " " else ""}\
${if cpu>1 then cpu else ""} \
        ${true="-l h_vmem=$(expr " false="" defined(memory_mb)}${memory_mb}\
${true=" / " false="" defined(memory_mb)}\
${if defined(memory_mb) then cpu else ""}\
${true=")m" false="" defined(memory_mb)} \
        ${true="-l s_vmem=$(expr " false="" defined(memory_mb)}${memory_mb}\
${true=" / " false="" defined(memory_mb)}\
${if defined(memory_mb) then cpu else ""}\
${true=")m" false="" defined(memory_mb)} \
        ${true="-l h_rt=" false="" defined(time)}${time}$\
{true=":00:00" false="" defined(time)} \
        ${true="-l s_rt=" false="" defined(time)}${time}$\
{true=":00:00" false="" defined(time)} \
        ${"-q " + sge_queue} \
        ${"-l gpu=" + gpu} \
        ${sge_extra_param} \
        -V
    """
    TEMPLATE = {
        "backend": {
            "providers": {
                BACKEND_SGE: {
                    "config": {
                        "default-runtime-attributes": {
                            "time": 24
                        },
                        "runtime-attributes": RUNTIME_ATTRIBUTES,
                        "submit": SUBMIT,
                        "exit-code-timeout-seconds": 180,
                        "kill": "qdel ${job_id}",
                        "check-alive": "qstat -j ${job_id}",
                        "job-id-regex": "(\\d+)"
                    }
                }
            }
        }
    }

    def __init__(self, pe=None, queue=None, extra_param=None):
        super().__init__(
            CaperBackendSGE.TEMPLATE,
            backend_name=BACKEND_SGE)
        config = self.get_backend_config()

        if pe is not None and pe != '':
            config['default-runtime-attributes']['sge_pe'] = pe
        if queue is not None and queue != '':
            config['default-runtime-attributes']['sge_queue'] = queue
        if extra_param is not None and extra_param != '':
            config['default-runtime-attributes']['sge_extra_param'] = extra_param


class CaperBackendPBS(CaperBackendBaseLocal):
    """PBS backend
    """
    RUNTIME_ATTRIBUTES = """
        String? docker
        String? docker_user
        Int cpu = 1
        Int? gpu
        Int? time
        Int? memory_mb
        String? pbs_queue
        String? pbs_extra_param
        String? singularity
        String? singularity_bindpath
        String? singularity_cachedir
    """
    SUBMIT = """
        echo "${if defined(singularity) then '' else '/bin/bash ${script} #'} \
        if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then \
        export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \
        if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then \
        export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi; \
        singularity exec --cleanenv --home ${cwd} \
        ${if defined(gpu) then '--nv' else ''} \
        ${singularity} /bin/bash ${script}" | qsub \
        -N ${job_name} \
        -o ${out} \
        -e ${err} \
        ${true="-lselect=1:ncpus=" false="" defined(cpu)}${cpu}\
${true=":mem=" false="" defined(memory_mb)}${memory_mb}\
${true="mb" false="" defined(memory_mb)} \
        ${true="-lwalltime=" false="" defined(time)}${time}\
${true=":0:0" false="" defined(time)} \
        ${true="-lngpus=" false="" gpu>1}${if gpu>1 then gpu else ""} \
        ${"-q " + pbs_queue} \
        ${pbs_extra_param} \
        -V
    """
    TEMPLATE = {
        "backend": {
            "providers": {
                BACKEND_PBS: {
                    "config": {
                        "default-runtime-attributes": {
                            "time": 24
                        },
                        "script-epilogue": "sleep 30 && sync",
                        "runtime-attributes": RUNTIME_ATTRIBUTES,
                        "submit": SUBMIT,
                        "exit-code-timeout-seconds": 180,
                        "kill": "qdel ${job_id}",
                        "check-alive": "qstat ${job_id}",
                        "job-id-regex": "(\\d+)"
                    }
                }
            }
        }
    }

    def __init__(self, queue=None, extra_param=None):
        super().__init__(
            CaperBackendPBS.TEMPLATE,
            backend_name=BACKEND_PBS)
        config = self.get_backend_config()

        if queue is not None and queue != '':
            config['default-runtime-attributes']['pbs_queue'] = queue
        if extra_param is not None and extra_param != '':
            config['default-runtime-attributes']['pbs_extra_param'] = extra_param


def main():
    pass


if __name__ == '__main__':
    main()



import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='caper',
    version='0.8.2',
    python_requires='>=3.6',
    scripts=['bin/caper', 'bin/run_mysql_server_docker.sh',
             'bin/run_mysql_server_singularity.sh'],
    author='Jin Lee',
    author_email='leepc12@gmail.com',
    description='Cromwell Assisted Pipeline ExecutoR',
    long_description='https://github.com/ENCODE-DCC/caper',  #long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ENCODE-DCC/caper',
    packages=setuptools.find_packages(exclude=['mysql*', 'docs']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
    ],
    install_requires=['pyhocon>=0.3.53', 'requests', 'pyopenssl', 'autouri>=0.1.2.1']
)



import os
import re
import shutil
from autouri import AutoURI, AbsPath, HTTPURL
from tempfile import TemporaryDirectory


class CaperWDLParser(object):
    """WDL parser for Caper.
    Find special comments for Caper in WDL.
    For example,
        #CAPER docker ubuntu:latest
    Find subworkflows and zip it.
    """
    RE_PATTERN_WDL_IMPORT = r'^\s*import\s+[\"\'](.+)[\"\']\s*'
    RE_PATTERN_WDL_COMMENT_DOCKER = r'^\s*\#\s*CAPER\s+docker\s(.+)'
    RE_PATTERN_WDL_COMMENT_SINGULARITY = r'^\s*\#\s*CAPER\s+singularity\s(.+)'
    RECURSION_DEPTH_LIMIT = 20

    def __init__(self, wdl):
        self._wdl = AbsPath.get_abspath_if_exists(wdl)

    def find_val(self, regex_val):
        u = AutoURI(self._wdl)
        if not u.exists:
            raise ValueError('WDL does not exist: wdl={wdl}'.format(wdl=self._wdl))
        result = []
        for line in u.read().split('\n'):
            r = re.findall(regex_val, line)
            if len(r) > 0:
                ret = r[0].strip()
                if len(ret) > 0:
                    result.append(ret)
        return result

    def find_imports(self):
        r = self.find_val(
            CaperWDLParser.RE_PATTERN_WDL_IMPORT)
        return r

    def find_docker(self):
        r = self.find_val(
            CaperWDLParser.RE_PATTERN_WDL_COMMENT_DOCKER)
        return r[0] if len(r) > 0 else None

    def find_singularity(self):
        r = self.find_val(
            CaperWDLParser.RE_PATTERN_WDL_COMMENT_SINGULARITY)
        return r[0] if len(r) > 0 else None

    def zip_subworkflows(self, zip_file):
        """Zip imported subworkflow WDLs (with relative paths only).
        For this (main) workflow, any URI is allowed.
        However, only subworkflows with relative path will be zipped.
        """
        with TemporaryDirectory() as tmp_d:
            main_wdl = AbsPath.localize(self._wdl)
            u = AutoURI(main_wdl)
            # with a directory structure as they imported
            num_sub_wf_packed = self.__recurse_subworkflow(
                root_wdl_dir=u.dirname,
                root_zip_dir=tmp_d)
            if num_sub_wf_packed:
                shutil.make_archive(AutoURI(zip_file).uri_wo_ext, 'zip', tmp_d)
            return num_sub_wf_packed

    def __recurse_subworkflow(
        self,
        root_zip_dir=None,
        root_wdl_dir=None,
        imported_as_url=False,
        depth=0):
        """Recurse imported sub-WDLs in main-WDL.

        Unlike Cromwell, Womtool does not take imports.zip while validating WDLs.
        All sub-WDLs should be in a correct directory structure relative to the 
        root WDL.
        For Womtool, we should make a temporary directory and unpack imports.zip there and
        need to make a copy of root WDL on it. Then run Womtool to validate them.
        This function is to make such imports.zip.

        Sub-WDLs imported as relative path simply inherit parent's directory.
        Sub-WDLs imported as URL does not inherit parent's directory but root 
        WDL's directory.
        Sub-WDLs imported as absolute path are not allowed. This can work with "caper run"
        but not with "caper submit" (or Cromwell submit).

        Args:
            depth: Recursion depth
        Returns:
            Total number of subworkflows:
                Sub WDL files "recursively" localized on "root_zip_dir".
        """
        if depth > CaperWDLParser.RECURSION_DEPTH_LIMIT:
            raise ValueError(
                'Reached recursion depth limit while zipping subworkflows recursively. '
                'Possible clyclic import or self-refencing in WDLs? wdl={wdl}'.format(
                    wdl=self._wdl))

        if imported_as_url:
            main_wdl_dir = root_wdl_dir
        else:
            main_wdl_dir = AbsPath(self._wdl).dirname

        num_sub_wf_packed = 0
        imports = self.find_imports()
        for sub_rel_to_parent in imports:
            u_sub = AutoURI(sub_rel_to_parent)

            if isinstance(u_sub, HTTPURL):
                sub_abs = u_sub.uri
                imported_as_url_sub = True
            elif isinstance(u_sub, AbsPath):
                raise ValueError(
                    'For sub WDL zipping, absolute path is not allowed for sub WDL. '
                    'main={main}, sub={sub}'.format(
                        main=self._wdl, sub=sub_rel_to_parent))
            else:
                sub_abs = os.path.realpath(
                    os.path.join(main_wdl_dir, sub_rel_to_parent))
                u_sub_abs = AbsPath(sub_abs)
                if not u_sub_abs.exists:
                    raise FileNotFoundError(
                        'Sub WDL does not exist. Did you import main WDL '
                        'as a URL but sub WDL references a local file? '
                        'main={main}, sub={sub}, imported_as_url={i}'.format(
                            main=self._wdl, sub=sub_rel_to_parent, i=imported_as_url))
                if not sub_abs.startswith(root_wdl_dir):
                    raise ValueError(
                        'Sub WDL exists but it is out of root WDL directory. '
                        'Too many "../" in your sub WDL? '
                        'Or main WDL is imported as an URL but sub WDL '
                        'has "../"? '
                        'main={main}, sub={sub}, imported_as_url={i}'.format(
                            main=self._wdl, sub=sub_rel_to_parent, i=imported_as_url))

                # make a copy on zip_dir
                rel_path = os.path.relpath(sub_abs, root_wdl_dir)
                cp_dest = os.path.join(root_zip_dir, rel_path)

                u_sub_abs.cp(cp_dest)
                num_sub_wf_packed += 1
                imported_as_url_sub = False

            num_sub_wf_packed += CaperWDLParser(sub_abs).__recurse_subworkflow(
                root_zip_dir=root_zip_dir,
                root_wdl_dir=root_wdl_dir,
                imported_as_url=imported_as_url_sub,
                depth=depth + 1)
        return num_sub_wf_packed



"""CaperArgs: Command line arguments parser for Caper

Author:
    Jin Lee (leepc12@gmail.com) at ENCODE-DCC
"""

import argparse
from configparser import ConfigParser
import sys
import os
from distutils.util import strtobool
from collections import OrderedDict
from .caper_backend import CaperBackendDatabase
from .caper_backend import CaperBackendGCP
from .caper_backend import BACKENDS, BACKEND_LOCAL
from .caper_backend import BACKEND_ALIAS_LOCAL
from .caper_backend import BACKEND_ALIAS_SHERLOCK, BACKEND_ALIAS_SCG


__version__ = '0.8.2'

DEFAULT_JAVA_HEAP_SERVER = '10G'
DEFAULT_JAVA_HEAP_RUN = '3G'
DEFAULT_CAPER_CONF = '~/.caper/default.conf'
DEFAULT_SINGULARITY_CACHEDIR = '~/.caper/singularity_cachedir'
DEFAULT_CROMWELL_JAR = 'https://github.com/broadinstitute/cromwell/releases/download/47/cromwell-47.jar'
DEFAULT_WOMTOOL_JAR = 'https://github.com/broadinstitute/cromwell/releases/download/47/womtool-47.jar'
DEFAULT_DB = CaperBackendDatabase.DB_TYPE_IN_MEMORY
DEFAULT_MYSQL_DB_IP = 'localhost'
DEFAULT_MYSQL_DB_PORT = 3306
DEFAULT_MYSQL_DB_USER = 'cromwell'
DEFAULT_MYSQL_DB_NAME = 'cromwell'
DEFAULT_MYSQL_DB_PASSWORD = 'cromwell'
DEFAULT_POSTGRESQL_DB_IP = 'localhost'
DEFAULT_POSTGRESQL_DB_PORT = 5432
DEFAULT_POSTGRESQL_DB_USER = 'cromwell'
DEFAULT_POSTGRESQL_DB_NAME = 'cromwell'
DEFAULT_POSTGRESQL_DB_PASSWORD = 'cromwell'
DEFAULT_DB_TIMEOUT_MS = 30000
DEFAULT_MAX_CONCURRENT_WORKFLOWS = 40
DEFAULT_MAX_CONCURRENT_TASKS = 1000
DEFAULT_MAX_RETRIES = 1
DEFAULT_PORT = 8000
DEFAULT_IP = 'localhost'
DEFAULT_FORMAT = 'id,status,name,str_label,user,submission'
DEFAULT_DEEPCOPY_EXT = 'json,tsv'
DEFAULT_SERVER_HEARTBEAT_FILE = '~/.caper/default_server_heartbeat'
DEFAULT_SERVER_HEARTBEAT_TIMEOUT_MS = 120000
DEFAULT_CONF_CONTENTS = '\n\n'
DEFAULT_GCP_CALL_CACHING_DUP_STRAT = CaperBackendGCP.CALL_CACHING_DUP_STRAT_REFERENCE

DYN_FLAGS = ['--singularity', '--docker']
INVALID_EXT_FOR_DYN_FLAG = '.wdl'


def process_dyn_flags(remaining_args, dyn_flags,
                      invalid_ext_for_dyn_flag):
    """Special treatment for dynamic flags which can be used as
    both params and flags

    Example1: caper run --docker atac.wdl
    atac.wdl can be misinterpreated as a docker image
    This function switches the order of --docker and atac.wdl
    Result1: caper run atac.wdl --docker

    Example2: caper run --singularity --docker atac.wdl
    This example switches twice. This is just for an example
    --singularity and --docker are mutually exclusive
    Result2: caper run atac.wdl --singularity --docker
    """
    for f in DYN_FLAGS:
        if f in remaining_args:
            loc = remaining_args.index(f)
            if loc < len(remaining_args) - 1:
                if remaining_args[loc + 1].endswith(INVALID_EXT_FOR_DYN_FLAG):
                    remaining_args[loc], remaining_args[loc + 1] = \
                        remaining_args[loc + 1], remaining_args[loc]
    return remaining_args


def parse_caper_arguments():
    """Argument parser for Caper
    """
    # write default conf file if not exists
    default_caper_conf = os.path.expanduser(DEFAULT_CAPER_CONF)
    if not os.path.exists(default_caper_conf):
        os.makedirs(os.path.dirname(default_caper_conf), exist_ok=True)
        with open(default_caper_conf, 'w') as fp:
            fp.write(DEFAULT_CONF_CONTENTS)

    conf_parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False)
    conf_parser.add_argument('-c', '--conf', help='Specify config file',
                             metavar='FILE',
                             default=DEFAULT_CAPER_CONF)
    conf_parser.add_argument('-v', '--version', action='store_true',
                             help='Show version')
    known_args, remaining_argv = conf_parser.parse_known_args()
    if known_args.version is not None and known_args.version:
        print(__version__)
        conf_parser.exit()
    process_dyn_flags(remaining_argv, DYN_FLAGS, INVALID_EXT_FOR_DYN_FLAG)

    # read conf file if it exists
    defaults = {}

    if known_args.conf is not None:
        # resolve tilde (~) in conf path
        known_args.conf = os.path.expanduser(known_args.conf)
        if os.path.exists(known_args.conf):
            config = ConfigParser()
            with open(known_args.conf, 'r') as fp:
                conf_contents = fp.read()
            if '[defaults]' not in conf_contents.split('\n'):
                conf_contents = '[defaults]\n' + conf_contents
            config.read_string(conf_contents)
            d = dict(config.items('defaults'))
            # remove keys with empty string
            d = {k: v.strip('"\'') for k, v in d.items() if v != ''}
            # replace - with _
            defaults.update({k.replace('-', '_'): v for k, v in d.items()})

    parser = argparse.ArgumentParser(parents=[conf_parser])
    subparser = parser.add_subparsers(dest='action')

    parent_init = argparse.ArgumentParser(add_help=False)
    choices = list(BACKENDS)
    choices.pop(choices.index(BACKEND_LOCAL))
    choices += [BACKEND_ALIAS_SHERLOCK, BACKEND_ALIAS_SCG, BACKEND_ALIAS_LOCAL]
    parent_init.add_argument('platform',
        choices=choices,
        help='Platform to initialize Caper for.')

    # all
    parent_all = argparse.ArgumentParser(add_help=False)
    parent_all.add_argument('--dry-run',
        action='store_true',
        help='Caper does not take any action.')

    group_log_level = parent_all.add_mutually_exclusive_group()
    group_log_level.add_argument('-V', '--verbose', action='store_true',
                   help='Prints all logs >= INFO level')
    group_log_level.add_argument('-D', '--debug', action='store_true',
                   help='Prints all logs >= DEBUG level')

    # run, server, submit
    parent_backend = argparse.ArgumentParser(add_help=False)
    parent_backend.add_argument(
        '-b', '--backend', help='Backend to run a workflow')

    # run, server
    parent_host = argparse.ArgumentParser(add_help=False)

    group_db = parent_host.add_argument_group(
        title='General DB settings (for both file DB and MySQL DB)')
    group_db.add_argument(
        '--db', choices=[
            CaperBackendDatabase.DB_TYPE_IN_MEMORY,
            CaperBackendDatabase.DB_TYPE_FILE,
            CaperBackendDatabase.DB_TYPE_MYSQL,
            CaperBackendDatabase.DB_TYPE_POSTGRESQL
        ],
        default=DEFAULT_DB,
        help='Cromwell metadata database type')
    group_db.add_argument(
        '--db-timeout', type=int, default=DEFAULT_DB_TIMEOUT_MS,
        help='Milliseconds to wait for DB connection.')

    group_file_db = parent_host.add_argument_group(
        title='HyperSQL file DB arguments (unstable, not recommended)')
    group_file_db.add_argument(
        '--file-db', '-d',
        help='Default DB file for Cromwell\'s built-in HyperSQL database.')

    group_mysql = parent_host.add_argument_group(
        title='MySQL DB arguments')
    group_mysql.add_argument(
        '--mysql-db-ip', default=DEFAULT_MYSQL_DB_IP,
        help='MySQL Database IP address (e.g. localhost)')
    group_mysql.add_argument(
        '--mysql-db-port', type=int, default=DEFAULT_MYSQL_DB_PORT,
        help='MySQL Database TCP/IP port (e.g. 3306)')
    group_mysql.add_argument(
        '--mysql-db-user', default=DEFAULT_MYSQL_DB_USER,
        help='MySQL DB username')
    group_mysql.add_argument(
        '--mysql-db-password', default=DEFAULT_MYSQL_DB_PASSWORD,
        help='MySQL DB password')
    group_mysql.add_argument(
        '--mysql-db-name', default=DEFAULT_MYSQL_DB_NAME,
        help='MySQL DB name for Cromwell')

    group_postgresql = parent_host.add_argument_group(
        title='PostgreSQL DB arguments')
    group_postgresql.add_argument(
        '--postgresql-db-ip', default=DEFAULT_POSTGRESQL_DB_IP,
        help='PostgreSQL DB IP address (e.g. localhost)')
    group_postgresql.add_argument(
        '--postgresql-db-port', type=int, default=DEFAULT_POSTGRESQL_DB_PORT,
        help='PostgreSQL DB TCP/IP port (e.g. 5432)')
    group_postgresql.add_argument(
        '--postgresql-db-user', default=DEFAULT_POSTGRESQL_DB_USER,
        help='PostgreSQL DB username')
    group_postgresql.add_argument(
        '--postgresql-db-password', default=DEFAULT_POSTGRESQL_DB_PASSWORD,
        help='PostgreSQL DB password')
    group_postgresql.add_argument(
        '--postgresql-db-name', default=DEFAULT_POSTGRESQL_DB_NAME,
        help='PostgreSQL DB name for Cromwell')

    group_cromwell = parent_host.add_argument_group(
        title='Cromwell settings')
    group_cromwell.add_argument(
        '--cromwell', default=DEFAULT_CROMWELL_JAR,
        help='Path or URL for Cromwell JAR file')
    group_cromwell.add_argument(
        '--max-concurrent-tasks', default=DEFAULT_MAX_CONCURRENT_TASKS,
        type=int,
        help='Number of concurrent tasks. '
             '"config.concurrent-job-limit" in Cromwell backend configuration '
             'for each backend')
    group_cromwell.add_argument(
        '--max-concurrent-workflows', default=DEFAULT_MAX_CONCURRENT_WORKFLOWS,
        type=int,
        help='Number of concurrent workflows. '
             '"system.max-concurrent-workflows" in backend configuration')
    group_cromwell.add_argument(
        '--max-retries', default=DEFAULT_MAX_RETRIES,
        type=int,
        help='Number of retries for failing tasks. '
             'equivalent to "maxRetries" in workflow options JSON file.')
    group_cromwell.add_argument(
        '--disable-call-caching', action='store_true',
        help='Disable Cromwell\'s call caching, which re-uses outputs from '
             'previous workflows')
    group_cromwell.add_argument(
        '--backend-file',
        help='Custom Cromwell backend configuration file to override all')
    group_cromwell.add_argument(
        '--soft-glob-output', action='store_true',
        help='Use soft-linking when globbing outputs for a filesystem that '
             'does not allow hard-linking. e.g. beeGFS. '
             'This flag does not work with backends based on a Docker container. '
             'i.e. gcp and aws. Also, '
             'it does not work with local backends (local/slurm/sge/pbs) '
             'with --docker. However, it works fine with --singularity.')

    group_local = parent_host.add_argument_group(
        title='local backend arguments')
    group_local.add_argument(
        '--out-dir', default='.', help='Output directory for local backend')
    group_local.add_argument(
        '--tmp-dir', help='Temporary directory for local backend')

    group_gc = parent_host.add_argument_group(
        title='GCP backend arguments')
    group_gc.add_argument('--gcp-prj', help='GC project')
    group_gc.add_argument('--gcp-zones', help='GCP zones (e.g. us-west1-b,'
                                              'us-central1-b)')
    group_gc.add_argument(
        '--gcp-call-caching-dup-strat', default=DEFAULT_GCP_CALL_CACHING_DUP_STRAT,
        choices=[
            CaperBackendGCP.CALL_CACHING_DUP_STRAT_REFERENCE,
            CaperBackendGCP.CALL_CACHING_DUP_STRAT_COPY
        ],
        help='Duplication strategy for call-cached outputs for GCP backend: '
             'copy: make a copy, reference: refer to old output in metadata.json.')
    group_gc.add_argument(
        '--out-gcs-bucket', help='Output GCS bucket for GCP backend')
    group_gc.add_argument(
        '--tmp-gcs-bucket', help='Temporary GCS bucket for GCP backend')

    group_aws = parent_host.add_argument_group(
        title='AWS backend arguments')
    group_aws.add_argument('--aws-batch-arn', help='ARN for AWS Batch')
    group_aws.add_argument('--aws-region', help='AWS region (e.g. us-west-1)')
    group_aws.add_argument(
        '--out-s3-bucket', help='Output S3 bucket for AWS backend')
    group_aws.add_argument(
        '--tmp-s3-bucket', help='Temporary S3 bucket for AWS backend')

    # run, submit
    parent_submit = argparse.ArgumentParser(add_help=False)

    parent_submit.add_argument(
        'wdl',
        help='Path, URL or URI for WDL script '
             'Example: /scratch/my.wdl, gs://some/where/our.wdl, '
             'http://hello.com/world/your.wdl')
    parent_submit.add_argument(
        '-i', '--inputs', help='Workflow inputs JSON file')
    parent_submit.add_argument(
        '-o', '--options', help='Workflow options JSON file')
    parent_submit.add_argument(
        '-l', '--labels',
        help='Workflow labels JSON file')
    parent_submit.add_argument(
        '-p', '--imports',
        help='Zip file of imported subworkflows')
    parent_submit.add_argument(
        '-s', '--str-label',
        help='Caper\'s special label for a workflow '
             'This label will be added to a workflow labels JSON file '
             'as a value for a key "caper-label". '
             'DO NOT USE IRREGULAR CHARACTERS. USE LETTERS, NUMBERS, '
             'DASHES AND UNDERSCORES ONLY (^[A-Za-z0-9\\-\\_]+$) '
             'since this label is used to compose a path for '
             'workflow\'s temporary directory (tmp_dir/label/timestamp/)')
    parent_submit.add_argument(
        '--hold', action='store_true',
        help='Put a hold on a workflow when submitted to a Cromwell server.')
    parent_submit.add_argument(
        '--singularity-cachedir', default=DEFAULT_SINGULARITY_CACHEDIR,
        help='Singularity cache directory. Equivalent to exporting an '
             'environment variable SINGULARITY_CACHEDIR. '
             'Define it to prevent repeatedly building a singularity image '
             'for every pipeline task')
    parent_submit.add_argument(
        '--use-gsutil-for-s3', action='store_true',
        help='Use gsutil CLI for direct trasnfer between S3 and GCS buckets. '
             'Otherwise, such file transfer will stream through local machine. '
             'Make sure that gsutil is installed on your system and it has access to '
             'credentials for AWS (e.g. ~/.boto or ~/.aws/credentials).')

    # server
    parent_server = argparse.ArgumentParser(add_help=False)
    parent_server.add_argument(
        '--java-heap-server', default=DEFAULT_JAVA_HEAP_SERVER,
        help='Cromwell Java heap size for "server" mode (java -Xmx)')

    # run
    parent_run = argparse.ArgumentParser(add_help=False)
    parent_run.add_argument(
        '-m', '--metadata-output',
        help='An optional directory path to output metadata JSON file')
    parent_run.add_argument(
        '--java-heap-run', default=DEFAULT_JAVA_HEAP_RUN,
        help='Cromwell Java heap size for "run" mode (java -Xmx)')

    parent_submit.add_argument(
        '--no-deepcopy', action='store_true',
        help='(IMPORTANT) --deepcopy has been deprecated. '
             'Deepcopying is now activated by default. '
             'This flag disables deepcopy for '
             'JSON (.json), TSV (.tsv) and CSV (.csv) '
             'files specified in an input JSON file (--inputs). '
             'Find all path/URI string values in an input JSON file '
             'and make copies of files on a local/remote storage '
             'for a target backend. Make sure that you have installed '
             'gsutil for GCS and aws for S3.')
    parent_submit.add_argument(
        '--ignore-womtool', action='store_true',
        help='Ignore warnings from womtool.jar.')
    parent_submit.add_argument(
        '--womtool', default=DEFAULT_WOMTOOL_JAR,
        help='Path or URL for Cromwell\'s womtool JAR file')

    group_dep = parent_submit.add_argument_group(
        title='dependency resolver for all backends',
        description=''
        'Cloud-based backends (gc and aws) will only use Docker '
        'so that "--docker URI_FOR_DOCKER_IMG" must be specified '
        'in the command line argument or as a comment "#CAPER '
        'docker URI_FOR_DOCKER_IMG" in a WDL file')
    group_dep.add_argument(
        '--docker', nargs='*',
        help='URI for Docker image (e.g. ubuntu:latest). '
        'This can also be used as a flag to use Docker image address '
        'defined in your WDL file as a comment ("#CAPER docker").')
    group_dep_local = parent_submit.add_argument_group(
        title='dependency resolver for local backend',
        description=''
        'Singularity is for local backend only. Other backends '
        '(gcp and aws) will use Docker only. '
        'Local backend defaults to not use any container-based methods. '
        'Use "--singularity" or "--docker" to use containers')
    group_dep_local.add_argument(
        '--singularity', nargs='*',
        help='URI or path for Singularity image '
             '(e.g. ~/.singularity/ubuntu-latest.simg, '
             'docker://ubuntu:latest, shub://vsoch/hello-world). '
             'This can also be used as a flag to use Docker image address '
             'defined in your WDL file as a comment ("#CAPER singularity").')
    group_dep_local.add_argument(
        '--no-build-singularity', action='store_true',
        help='Do not build singularity image before running a workflow. ')

    group_slurm = parent_submit.add_argument_group('SLURM arguments')
    group_slurm.add_argument('--slurm-partition', help='SLURM partition')
    group_slurm.add_argument('--slurm-account', help='SLURM account')
    group_slurm.add_argument(
        '--slurm-extra-param',
        help='SLURM extra parameters. Must be double-quoted')

    group_sge = parent_submit.add_argument_group('SGE arguments')
    group_sge.add_argument(
        '--sge-pe', help='SGE parallel environment. Check with "qconf -spl"')
    group_sge.add_argument(
        '--sge-queue', help='SGE queue. Check with "qconf -sql"')
    group_sge.add_argument(
        '--sge-extra-param',
        help='SGE extra parameters. Must be double-quoted')

    group_pbs = parent_submit.add_argument_group('PBS arguments')
    group_pbs.add_argument(
        '--pbs-queue', help='PBS queue')
    group_pbs.add_argument(
        '--pbs-extra-param',
        help='PBS extra parameters. Must be double-quoted')

    # list, metadata, abort
    parent_search_wf = argparse.ArgumentParser(add_help=False)
    parent_search_wf.add_argument(
        'wf_id_or_label', nargs='*',
        help='List of workflow IDs to find matching workflows to '
             'commit a specified action (list, metadata and abort). '
             'Wildcards (* and ?) are allowed.')

    parent_server_client = argparse.ArgumentParser(add_help=False)
    parent_server_client.add_argument(
        '--port', default=DEFAULT_PORT, type=int,
        help='Port for Caper server')
    parent_server_client.add_argument(
        '--ip', default=DEFAULT_IP,
        help='IP address for Caper server')
    parent_server_client.add_argument(
        '--no-server-heartbeat', action='store_true',
        help='Disable server heartbeat file.')
    parent_server_client.add_argument(
        '--server-heartbeat-file',
        default=DEFAULT_SERVER_HEARTBEAT_FILE,
        help='Heartbeat file for Caper clients to get IP and port of a server')
    parent_server_client.add_argument(
        '--server-heartbeat-timeout',
        default=DEFAULT_SERVER_HEARTBEAT_TIMEOUT_MS,
        help='Timeout for a heartbeat file in Milliseconds. '
             'A heartbeat file older than '
             'this interval will be ignored.')

    parent_list = argparse.ArgumentParser(add_help=False)
    parent_list.add_argument(
        '-f', '--format', default=DEFAULT_FORMAT,
        help='Comma-separated list of items to be shown for "list" '
        'subcommand. Any key name in workflow JSON from Cromwell '
        'server\'s response is allowed. '
        'Available keys are "id" (workflow ID), "status", "str_label", '
        '"name" (WDL/CWL name), "submission" (date/time), "start", '
        '"end" and "user". '
        '"str_label" is a special key for Caper. See help context '
        'of "--str-label" for details')
    parent_list.add_argument(
        '--hide-result-before',
        help='Hide workflows submitted before this date/time. '
             'Use the same (or shorter) date/time format shown in '
             '"caper list". '
             'e.g. 2019-06-13, 2019-06-13T10:07')
    # troubleshoot
    parent_troubleshoot = argparse.ArgumentParser(add_help=False)
    parent_troubleshoot.add_argument(
        '--show-completed-task', action='store_true',
        help='Show information about completed tasks.')

    p_init = subparser.add_parser(
        'init',
        help='Initialize Caper\'s configuration file. THIS WILL OVERWRITE ON '
             'A SPECIFIED(-c)/DEFAULT CONF FILE. e.g. ~/.caper/default.conf.',
        parents=[parent_init])
    p_run = subparser.add_parser(
        'run', help='Run a single workflow without server',
        parents=[parent_all, parent_submit, parent_run, parent_host, parent_backend])
    p_server = subparser.add_parser(
        'server', help='Run a Cromwell server',
        parents=[parent_all, parent_server_client, parent_server, parent_host,
                 parent_backend])
    p_submit = subparser.add_parser(
        'submit', help='Submit a workflow to a Cromwell server',
        parents=[parent_all, parent_server_client, parent_submit,
                 parent_backend])
    p_abort = subparser.add_parser(
        'abort', help='Abort running/pending workflows on a Cromwell server',
        parents=[parent_all, parent_server_client, parent_search_wf])
    p_unhold = subparser.add_parser(
        'unhold', help='Release hold of workflows on a Cromwell server',
        parents=[parent_all, parent_server_client, parent_search_wf])
    p_list = subparser.add_parser(
        'list', help='List running/pending workflows on a Cromwell server',
        parents=[parent_all, parent_server_client, parent_search_wf,
                 parent_list])
    p_metadata = subparser.add_parser(
        'metadata',
        help='Retrieve metadata JSON for workflows from a Cromwell server',
        parents=[parent_all, parent_server_client, parent_search_wf])
    p_troubleshoot = subparser.add_parser(
        'troubleshoot',
        help='Troubleshoot workflow problems from metadata JSON file or '
             'workflow IDs',
        parents=[parent_all, parent_troubleshoot, parent_server_client, parent_search_wf])
    p_debug = subparser.add_parser(
        'debug',
        help='Identical to "troubleshoot"',
        parents=[parent_all, parent_troubleshoot, parent_server_client, parent_search_wf])

    for p in [p_init, p_run, p_server, p_submit, p_abort, p_unhold, p_list,
              p_metadata, p_troubleshoot, p_debug]:
        p.set_defaults(**defaults)

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    # parse all args
    args = parser.parse_args(remaining_argv)

    # convert to dict
    args_d = vars(args)

    # string to boolean
    for k in [
        'dry_run',
        'no_server_heartbeat',
        'disable_call_caching',
        'soft_glob_output',
        'hold',
        'no_deepcopy',
        'ignore_womtool',
        'no_build_singularity',
        'use_gsutil_for_s3',
        'show_completed_task']:
        v = args_d.get(k)
        if v is not None and isinstance(v, str):
            args_d[k] = bool(strtobool(v))

    # string to int
    for k in [
        'db_timeout',
        'max_retries',
        'max_concurrent_tasks',
        'max_concurrent_workflows',
        'mysql_db_port',
        'postgresql_db_port',
        'server_heartbeat_timeout',
        'port']:
        v = args_d.get(k)
        if v is not None and isinstance(v, str):
            args_d[k] = int(v)

    return args_d
